import copy
import torch
from .._gp import update_gp, predict, predictive_covariance
from .._utils import Utils


class ScaleVanillaGP:
    def __init__(
        self,
        Xobs,
        Yobs,
        gp_kernel,
        device,
        lik=1e-10,
        training_iter=10000,
        thresh=0.01,
        lr=0.1,
        rng=10,
        train_lik=False,
        optimiser="BoTorch",
    ):
        """
        Vanilla Gaussian process modelling for Bayesian quadrature.
        The observed Y (Yobs) is assumed to be non-normalised log-likelihood.
        "Scale" here means this model automatically normalise the log-likelihood for vanilla GP.
        
        Args:
           - Xobs: torch.tensor, X samples, X belongs to prior measure.
           - Yobs: torch.tensor, Y observations, Y = true_likelihood(X).
           - gp_kernel: gpytorch.kernels, GP kernel function
           - device: torch.device, device, cpu or cuda
           - lik: float, the initial value of GP likelihood noise variance
           - train_iter: int, the maximum iteration for GP hyperparameter training.
           - thresh: float, the threshold as a stopping criterion of GP hyperparameter training.
           - lr: float, the learning rate of Adam optimiser
           - rng: int, tne range coefficient of GP likelihood noise variance
           - train_like: bool, flag whether or not to update GP likelihood noise variance
        """
        self.gp_kernel = gp_kernel
        self.device = device
        self.lik = lik
        self.training_iter = training_iter
        self.thresh = thresh
        self.lr = lr
        self.rng = rng
        self.train_lik = train_lik
        self.optimiser = optimiser

        self.jitter = 1e-6
        self.Y_log = copy.deepcopy(Yobs)
        self.utils = Utils(device)

        self.model = update_gp(
            Xobs,
            self.process_y_with_scaling(Yobs),
            gp_kernel,
            self.device,
            lik=self.lik,
            training_iter=self.training_iter,
            thresh=self.thresh,
            lr=self.lr,
            rng=self.rng,
            train_lik=self.train_lik,
            optimiser=self.optimiser,
        )
    
    def process_y_with_scaling(self, y_obs):
        """
        Args:
           - y_obs: torch.tensor, observations of true_loglikelihood

        Returns:
           - y_h: torch.tensor, warped observations in h space that contains no anomalies and the updated alpha hyperparameter.
        """

        y = self.utils.remove_anomalies(y_obs)
        self.beta = torch.max(y)
        y_exp = torch.exp(y - self.beta)
        return y_exp

    def cat_observations_with_scaling(self, X, Y):
        """
        Args:
           - X: torch.tensor, X samples to be added to the existing data Xobs
           - Y: torch.tensor, unwarped Y observations to be added to the existing data Yobs

        Returns:
           - Xall: torch.tensor, X samples that contains all samples
           - Yall: torch.tensor, warped Y observations that contains all observations
        """
        Xobs = self.model.train_inputs[0]
        Yobs_log = copy.deepcopy(self.Y_log)
        if len(self.model.train_targets.shape) == 0:
            Yobs_log = Yobs_log.unsqueeze(0)
        Xall = torch.cat([Xobs, X])
        Yall_log = torch.cat([Yobs_log, Y])
        self.Y_log = copy.deepcopy(Yall_log)
        Yall_exp = self.process_y_with_scaling(Yall_log)
        return Xall, Yall_exp

    def update_gp(self, X, Y):
        """
        Args:
           - X: torch.tensor, X samples to be added to the existing data Xobs
           - Y: torch.tensor, Y observations to be added to the existing data Yobs
        """
        Xall, Yall = self.cat_observations_with_scaling(X, Y)
        self.model = update_gp(
            Xall,
            Yall,
            self.gp_kernel,
            self.device,
            lik=self.lik,
            training_iter=self.training_iter,
            thresh=self.thresh,
            lr=self.lr,
            rng=self.rng,
            train_lik=self.train_lik,
            optimiser=self.optimiser,
        )

    def retrain_gp(self):
        Xobs = self.model.train_inputs[0]
        Yobs = self.model.train_targets
        self.model = update_gp(
            Xobs,
            Yobs,
            self.gp_kernel,
            self.device,
            lik=self.lik,
            training_iter=self.training_iter,
            thresh=self.thresh,
            lr=self.lr,
            rng=self.rng,
            train_lik=self.train_lik,
            optimiser=self.optimiser,
        )

    def predictive_kernel(self, x, y):
        """
        Args:
           - x: torch.tensor, x locations to be predicted
           - y: torch.tensor, y locations to be predicted

        Args:
           - CLy: torch.tensor, the positive semi-definite Gram matrix of predictive variance
        """
        return predictive_covariance(x, y, self.model)

    def predict(self, x):
        """
        Args:
           - x: torch.tensor, x locations to be predicted

        Returns:
           - mu: torch.tensor, predictive mean at given locations x.
           - var: torch.tensor, predictive variance at given locations x.
        """
        mu, var = predict(x, self.model)
        return mu, var

    def predict_mean(self, x):
        """
        Args:
           - x: torch.tensor, x locations to be predicted

        Returns:
           - mu: torch.tensor, predictive mean at given locations x.
        """
        mu, _ = predict(x, self.model)
        return mu



import torch
import torch.optim as optim
import torch.distributions as D
from abc import ABC, abstractmethod
from ._wkde import WeightedKernelDensityEstimation
from ._utils import TensorManager


class BaseMLE(ABC, TensorManager):
    def __init__(self):
        super().__init__()
    
    @abstractmethod
    def objective(self, X):
        r"""Optimization objective (weights)"""
        pass
    
    @abstractmethod
    def closure(self, X):
        r"""Optimization loop"""
        pass
    
    @abstractmethod
    def run(self, X):
        r"""Running optimization loop"""
        pass
    
    @abstractmethod
    def update_prior(self, X):
        r"""Return the updated prior"""
        pass

class BernoulliMLE(BaseMLE):
    def __init__(self, weights, x_binary, n_max=5):
        """
        Update Bernoulli prior via maximum likelihood estimation (MLE).
        
        Args:
        - weights: torch.tensor, the weights at the observed input
        - x_binary: torch.tensor, the observed input
        - n_max: int, the number of L-BFGS-B iteration
        """
        super().__init__() # call TensorManager
        self.weights = weights.detach()
        self.x_binary = x_binary
        self.n_dims_binary = x_binary.size(1)
        self.n_max = n_max
    
    def objective(self, w):
        """
        The objective of L-BFGS-B loop (maximum likelihood)
        
        Args:
        - w: torch.tensor, the weights to be optimised
        
        Return:
        - ans: torch.float, the negative log likelihood of the given w
        """
        dist = D.Bernoulli(w)
        ans = self.weights @ dist.log_prob(self.x_binary).sum(axis=1)
        return -1 * ans
    
    def transform(self, w):
        """
        Sigmoid transform to make w to be bounded from 0 to 1
        
        Args:
        - w: torch.tensor, the weights to be optimised
        
        Return:
        - w_trans: torch.tensor, the transformed weights
        """
        return 1/(1 + w.exp())

    def closure(self):
        """
        A single step closure of iteration loop
        
        Return:
        objective: torch.float, the negative log likelihood of the given w
        """
        self.lbfgs.zero_grad()
        params = self.transform(self.x_lbfgs)
        objective = self.objective(params)
        objective.backward()
        return objective
    
    def run(self):
        """
        Maximum likelihood estimation of optimal weights for the Bernoulli sampler
        
        Return:
        result: torch.tensor, the optimised weights for the Bernoulli sampler
        """
        self.x_lbfgs = self.ones(self.n_dims_binary) * 0.5
        self.x_lbfgs.requires_grad = True
        
        self.lbfgs = optim.LBFGS(
            [self.x_lbfgs],
            history_size=10, 
            max_iter=4, 
            line_search_fn="strong_wolfe",
        )
        
        for i in range(self.n_max):
            self.lbfgs.step(self.closure)
        result = self.transform(self.x_lbfgs).detach()
        return result
    
    def update_prior(self, prior_binary):
        """
        Update the Bernoulli prior
        
        Args:
        - prior_binary: torch.distributions.Bernoulli, the Bernoulli prior
        
        Return:
        - prior_binary: torch.distributions.Bernoulli, the Bernoulli prior
        """
        weights_updated = self.run()
        prior_binary.probs = weights_updated
        return prior_binary
    
class CategoricalMLE(BaseMLE):
    def __init__(self, weights, x_disc, prior, n_max=5):
        """
        Update Bernoulli prior via maximum likelihood estimation (MLE).
        
        Args:
        - weights: torch.tensor, the weights at the observed input
        - x_disc: torch.tensor, the observed input
        - prior: class, the function of categorical prior
        - n_max: int, the number of L-BFGS-B iteration
        """
        super().__init__() # call TensorManager
        self.weights = weights.detach()
        self.x_disc = x_disc.detach()
        self.n_dims_disc = x_disc.size(1)
        self.prior = prior
        self.n_max = n_max
        
    def reshape_weights(self, _w):
        """
        Reshape the weights for CategoricalPrior class
        
        Args:
        - _w: torch.tensor, the one-dimensional weights to be optimised
        
        Return:
        - w: list, the reshaped weights _w
        """
        sections = torch.cat([torch.tensor([0]), self.prior.n_categories.cumsum(0)])
        w = [_w[sections[idx]:sections[idx+1]] for idx in range(self.prior.n_dims)]
        return w
    
    def objective(self, _w):
        """
        The objective of L-BFGS-B loop (maximum likelihood)
        
        Args:
        - w: torch.tensor, the weights to be optimised
        
        Return:
        - ans: torch.float, the negative log likelihood of the given w
        """
        w = self.reshape_weights(_w)
        self.prior.weights = w
        self.prior.initialise()
        ans = self.weights @ self.prior.logpdf(self.x_disc)
        return -1 * ans
    
    def transform(self, w):
        """
        Sigmoid transform to make w to be bounded from 0 to 1
        
        Args:
        - w: torch.tensor, the weights to be optimised
        
        Return:
        - w_trans: torch.tensor, the transformed weights
        """
        return 1/(1 + w.exp())
    
    def closure(self):
        """
        A single step closure of iteration loop
        
        Return:
        objective: torch.float, the negative log likelihood of the given w
        """
        self.lbfgs.zero_grad()
        params = self.transform(self.x_lbfgs)
        objective = self.objective(params)
        objective.backward(retain_graph=True)
        return objective
    
    def run(self):
        """
        Maximum likelihood estimation of optimal weights for the categorical sampler
        
        Return:
        result: torch.tensor, the optimised weights for the Bernoulli sampler
        """
        self.x_lbfgs = torch.hstack(self.prior.weights)
        self.x_lbfgs.requires_grad = True
        
        self.lbfgs = optim.LBFGS([self.x_lbfgs],
                    history_size=10,
                    max_iter=4, 
                    line_search_fn="strong_wolfe")
                    
        for i in range(self.n_max):
            self.lbfgs.step(self.closure)
        result = self.transform(self.x_lbfgs).detach()
        return result
    
    def update_prior(self, prior_disc):
        """
        Update the categorical prior
        
        Args:
        - prior_disc: class, the categorical prior
        
        Return:
        - prior_disc: class, the optimised categorical prior
        """
        weights_updated = self.run()
        prior_disc.weights = self.reshape_weights(weights_updated)
        return prior_disc
    
def update_binary_prior(weights, x_binary, prior_binary):
    """
    Update the Bernoulli prior

    Args:
    - weights: torch.tensor, the weghts at X_cand
    - X_binary: torch.tensor, the binary input
    - prior_binary: torch.distributions.Bernoulli, the Bernoulli prior

    Return:
    - prior_binary: torch.distributions.Bernoulli, the Bernoulli prior
    """
    mle_binary = BernoulliMLE(weights, x_binary)
    prior_binary = mle_binary.update_prior(prior_binary)
    return prior_binary

def update_categorical_prior(weights, x_disc, prior_categorical):
    """
    Update the categorical prior

    Args:
    - weights: torch.tensor, the weghts at X_cand
    - X_disc: torch.tensor, the categorical input
    - prior_categorical: torch.distributions.Categorical, the Categorical prior

    Return:
    - prior_categorical: torch.distributions.Categorical, the optimised Categorical prior
    """
    mle_disc = CategoricalMLE(weights, x_disc, prior_categorical)
    prior_categorical =  mle_disc.update_prior(prior_categorical)
    return prior_categorical

def update_continuous_prior(X_cand, weights, prior, n_dims):
    """
    Update the continuous prior

    Args:
    - X_cand: torch.tensor, the mixed input
    - weights: torch.tensor, the weghts at X_cand
    - prior: class, the continuous prior
    - n_dims: int, the number of dimensions

    Return:
    - prior: class, the Gaussian mixture prior
    """
    if hasattr(prior, "bounds"):
        bounds = prior.bounds
    else:
        bounds = None
        
    prior = WeightedKernelDensityEstimation(
        X_cand, weights, n_dims, bounds=bounds,
    )
    return prior

def update_mixed_prior(X_cand, weights, prior, label="binary"):
    """
    Update the mixed prior

    Args:
    - X_cand: torch.tensor, the mixed input
    - weights: torch.tensor, the weghts at X_cand
    - prior: class, the mixed prior
    - label: string, "binary" or "categorical"

    Return:
    - prior: class, the mixed prior
    """
    x_cont, x_disc = prior.separate_samples(X_cand)
    if label == "binary":
        prior.prior_binary.prior_binary = update_binary_prior(
            weights, x_disc, prior.prior_binary.prior_binary,
        )
    elif label == "categorical":
        prior.prior_disc = update_categorical_prior(
            weights, x_disc, prior.prior_disc,
        )
    else:
        raise ValueError("label should be either 'binary' or 'categorical'.")
    
    prior.prior_cont = update_continuous_prior(x_cont, weights, prior, prior.n_dims_cont)
    return prior




import torch
import warnings
from torch.quasirandom import SobolEngine
from torch.distributions.multivariate_normal import MultivariateNormal
from ._settings import setting_parameters


def device_manager(device=None):
    if device == None:
        device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
    return device

def dtype_manager(dtype=None):
    if dtype == None:
        dtype = torch.float
    return dtype

class TensorManager:
    def __init__(self, device=None, dtype=None):
        _device, _dtype = setting_parameters()
        if dtype == None:
            dtype = _dtype
        if device == None:
            device = _device
        
        self.device = device_manager(device=device)
        self.dtype = dtype_manager(dtype=dtype)
        
    def standardise_tensor(self, tensor):
        return tensor.to(self.device, self.dtype)
    
    def standardise_device(self, tensor):
        return tensor.to(self.device)
    
    def ones(self, n_samples, n_dims=None):
        if n_dims == None:
            return self.standardise_tensor(torch.ones(n_samples))
        else:
            return self.standardise_tensor(torch.ones(n_samples, n_dims))
    
    def zeros(self, n_samples, n_dims=None):
        if n_dims == None:
            return self.standardise_tensor(torch.zeros(n_samples))
        else:
            return self.standardise_tensor(torch.zeros(n_samples, n_dims))
    
    def rand(self, n_dims, n_samples, qmc=True):
        if qmc:
            random_samples = SobolEngine(n_dims, scramble=True).draw(n_samples)
        else:
            random_samples = torch.rand(n_samples, n_dims)
        return self.standardise_tensor(random_samples)
    
    def arange(self, length):
        return self.standardise_device(torch.arange(length))
    
    def null(self):
        return self.standardise_device(torch.tensor([]))
    
    def tensor(self, x):
        return self.standardise_tensor(torch.tensor(x))
    
    def randperm(self, length):
        return self.standardise_device(torch.randperm(length))
    
    def multinomial(self, weights, n):
        return self.standardise_device(torch.multinomial(weights, n))
    
    def numpy(self, x):
        return x.detach().cpu().numpy()
    
    def is_cuda(self):
        if self.device == torch.device('cuda'):
            return True
        else:
            return False

class SafeTensorOperator(TensorManager):
    def __init__(self):
        super().__init__()
        self.eps = -torch.sqrt(torch.tensor(torch.finfo().max)).item()
        self.gpu_lim = int(5e5)
        self.max_iter = 10
        
    def remove_anomalies(self, y):
        """
        Args:
           - y: torch.tensor, observations

        Returns:
           - y: torch.tensor, observations whose anomalies have been removed.
        """
        y[y.isnan()] = self.eps
        y[y.isinf()] = self.eps
        y[y < self.eps] = self.eps
        return y

    def remove_anomalies_uniform(self, X, uni_min, uni_max):
        """
        Args:
           - X: torch.tensor, inputs
           - uni_min: torch.tensor, the minimum limit values of uniform distribution
           - uni_max: torch.tensor, the maximum limit values of uniform distribution

        Returns:
           - idx: bool, indices where the inputs X do not exceed the min-max limits
        """
        logic = torch.sum(torch.stack([torch.logical_or(
            X[:, i] < uni_min[i],
            X[:, i] > uni_max[i],
        ) for i in range(X.size(1))]), axis=0)
        return (logic == 0)

    def is_psd(self, mat):
        """
        Args:
           - mat: torch.tensor, symmetric matrix

        Returns:
           - flag: bool, flag to judge whether or not the given matrix is positive semi-definite
        """
        try:
            torch.linalg.cholesky(mat)
            return bool((mat == mat.T).all() and (torch.linalg.eig(mat)[0].real >= 0).all())
        except:
            return False
        
    def make_cov_psd(self, cov):
        """
        Args:
           - cov: torch.tensor, covariance matrix of multivariate normal distribution

        Returns:
           - cov: torch.tensor, covariance matrix of multivariate normal distribution
        """
        if self.is_psd(cov):
            return cov
        else:
            warnings.warn("Estimated covariance matrix was not positive semi-definite. Conveting...")
            cov = torch.nan_to_num(cov)
            cov = torch.sqrt(cov * cov.T)
            if not self.is_psd(cov):
                n_dim = cov.size(0)
                r_increment = 2
                jitter = self.ones(n_dim) * 1e-5
                n_iter = 0
                while not self.is_psd(cov):
                    cov[range(n_dim), range(n_dim)] += jitter
                    jitter *= r_increment
                    n_iter += 1
                    if n_iter > self.max_iter:
                        cov = cov.diag().diag()
                        break
            return cov

    def safe_mvn_register(self, mu, cov):
        """
        Args:
           - mu: torch.tensor, mean vector of multivariate normal distribution
           - cov: torch.tensor, covariance matrix of multivariate normal distribution

        Returns:
           - mvn: torch.distributions, function of multivariate normal distribution
        """
        cov = self.make_cov_psd(cov)
        return MultivariateNormal(mu, cov)
        
    def safe_mvn_prob(self, mu, cov, X):
        """
        Args:
           - mu: torch.tensor, mean vector of multivariate normal distribution
           - cov: torch.tensor, covariance matrix of multivariate normal distribution
           - X: torch.tensor, the locations that we wish to calculate the probability density values

        Returns:
           - pdf: torch.tensor, the probability density values at given locations X.
        """
        mvn = self.safe_mvn_register(mu, cov)
        if X.size(0) > self.gpu_lim:
            warnings.warn("The matrix size exceeds the GPU limit. Splitting.")
            n_split = torch.tensor(X.size(0) / self.gpu_lim).ceil().long()
            _X = torch.tensor_split(X, n_split)
            Npdfs = torch.cat(
                [
                    mvn.log_prob(_X[i]).exp()
                    for i in range(n_split)
                ]
            )
        else:
            Npdfs = mvn.log_prob(X).exp()
        return Npdfs
        
class Utils(SafeTensorOperator):
    def __init__(self):
        super().__init__()

    



import time
import copy
import torch
import warnings
from ._sampler import EmpiricalSampler
from ._kernel import Kernel
from ._pi import PI, PI_FBGP, PI_BQ

class Sober(EmpiricalSampler):
    def __init__(
        self,
        prior,
        model,
        thresh=5,
        sampler_type="lfi",
        kernel_type="predictive_covariance",
        dataset_pruning=True,
    ):
        """
        Batch Bayesian optimisation as batch Bayesian quadrature
        
        Args:
        - prior: class, the class of prior distribution
        - model: gpytorch.models, function of GP model.
        - eps: float, the machine epsilon (the smallest number of floating point).
               For double precision; eps = torch.finfo().min
        - thresh: int, the number of non-zero weights which regrads anomalies.
        - sampler_type: string, Select from "lfi" or "ts".
                        LFI = likelihood-free inference, TS = Thompson sampling
        - kernel_type: string, Select from ["predictive_covariance", "weighted_predictive_covariance"]
        - dataset_pruning: bool, perform pruning for dataset prior if true, otherwise not.
        """
        self.sampler_type = sampler_type
        self.kernel_type = kernel_type
        self.dataset_pruning = dataset_pruning
        self.check_model_type(model)
        pi, kernel = self.initialisation(model)
        self.n_batches_until_reset = 3
        super().__init__(prior, pi, kernel, label=prior.type)  # EmpiricalSampler class initialisation
    
    def check_model_type(self, model):
        # check fully Bayesian GP model or not
        if hasattr(model, "is_fbgp"):
            self.fbgp = True
            self.is_bq = False
            self.n_init = len(model.fobs)
        elif hasattr(model, "is_bq"):
            self.fbgp = False
            self.is_bq = True
            self.n_init = len(model.Y_log)
        else:
            self.fbgp = False
            self.is_bq = False
            self.n_init = len(model.train_targets)            
        
    def initialisation(self, model):
        """
        Set pi and kernel
        
        Args:
        - model: gpytorch.models, function of GP model.
        """
        if self.fbgp:
            pi = PI_FBGP(model)
            kernel = model.marginal_predictive_covariance
        elif self.is_bq:
            pi = PI_BQ(model)
            kernel = model.gspace_kernel
        else:
            pi = PI(model, label=self.sampler_type)
            kernel = Kernel(model, mode=self.kernel_type)
        return pi, kernel
    
    def update_model(self, model):
        """
        Set pi and kernel
        
        Args:
        - model: gpytorch.models, function of GP model.
        """
        pi, kernel = self.initialisation(model)
        super().__init__(self.prior, pi, kernel, thresh=self.thresh, label=self.prior.type)
    
    def should_reset_prior(self, batch_size, recycle_prior):
        """
        Check whether or not the prior should reset
        
        Args:
        - batch_size: int, the number of batch samples
        - recycle_prior: bool, recycle the previous prior if true, otherwise not.
        
        Return:
        - flag: bool, the prior should reset if true, otherwise not.
        """
        if self.fbgp:
            targets = self.pi.model.fobs
        elif self.is_bq:
            targets = self.pi.model.Y_log
        else:
            targets = self.pi.model.train_targets
        
        n_targets = len(targets)
        y_max = targets.max()
        cummax = targets.cummax(0).values
        learning_length = n_targets - self.n_init
        if (learning_length == 0) or (learning_length == batch_size):
            return False
        try:
            idx_max = torch.where((cummax >= y_max).diff() == True)[0][0]
        except:
            idx_max = self.tensor(0)
        n_interations = torch.tensor(learning_length / batch_size).ceil().long()
        for n_batches in range(1, n_interations+1):
            idx_batch = n_batches * batch_size
            if idx_batch >= idx_max:
                break
        n_nonimproved_batches = n_interations - n_batches + 2
        if n_nonimproved_batches >= self.n_batches_until_reset:
            return True
        elif not recycle_prior:
            return True
        else:
            return False
    
    def next_batch(
        self, 
        n_rec, 
        n_nys,
        batch_size, 
        calc_obj=None, 
        return_weights=False,
        recycle_prior=True,
        verbose=False,
    ):
        """
        Sampling the next batch location via kernel recombination.
        
        Args:
        - n_rec: int, the number of samples for recombination
        - n_nys: int, the number of samples for Nyström approximation
        - batch_size: int, the number of batch samples
        - calc_obj: class, the acquisition function (AF). Do not use AF if None.
        - return_weights: bool, return quadrature weights if true, otherwise not.
        - recycle_prior: bool, recycle the previous prior if true, otherwise not.
        - verbose: bool, show progress if truem otherwise not.
        
        Return:
        - X_batch: torch.tensor, the next batch samples
        """
        if verbose:
            start = time.monotonic()
            print("--- generating the candidates from pi...")
        if not self.label == "dataset":
            if self.should_reset_prior(batch_size, recycle_prior):
                print("The prior was initialised.")
                self.initialise_prior()
            X_cand, X_nys, weights = self.sampling_candidates(n_rec, n_nys, verbose=verbose)
        else:
            empirical_measure = self.sampling_datasets(n_rec, n_nys)
            if self.dataset_pruning:
                idx_sampled, X_cand, X_nys, weights = empirical_measure
            else:
                X_cand, X_nys, weights = empirical_measure
            
        if verbose:
            intermidiate = time.monotonic()
            print(f"--- Finished {intermidiate - start:.3e} [s]")
            print("|| summary of sampling ||")
            print(f" # of recombination samples: {len(X_cand):.3e}")
            print(f" # of Nyström samples: {len(X_nys):.3e}")
            print(f" # of nonzero weights: {(weights > 0).sum():.3e}")
            print("--- Start kernel recombination...")
        
        idx_rchq, w_rchq = self.sampling_recombination(
            X_cand,
            X_nys,
            weights,
            batch_size,
            calc_obj=calc_obj,
        )
        X_batch = X_cand[idx_rchq]
        if verbose:
            end = time.monotonic()
            print(f"--- Finished all tasks {end - start:.3e} [s]")
            
        if return_weights:
            return w_rchq, X_batch
        elif self.label == "dataset":
            if self.dataset_pruning:
                idx_rchq = idx_sampled[idx_rchq]
                return idx_rchq, X_batch
            else:
                return idx_rchq, X_batch
        else:
            return X_batch



from ._gp import predict
import torch.distributions as D


class PI:
    def __init__(self, model, label="lfi"):
        """
        Definition of pi (feasible resion).
        Select TS (Thompson Sampling) or LFI (likelihood-free inference)
        
        Args:
        - model: gpytorch.models, function of GP model
        - label: string, "ts" or "lfi"
        """
        self.model = model
        self.Xobs = model.train_inputs[0]    # training data X_obs
        self.eta = self.model.likelihood(self.model(self.Xobs)).loc.max().item() # current maximum
        self.label = label
    
    def lfi(self, X_cand, log=False):
        """
        Compute LFI at given locations X_cand
        
        Args:
        - X_cand: torch.tensor, inputs X where to compute LFI
        - log: bool: return log value if true, otherwise not.
        
        Return:
        - lfi: torch.tensor, the LFI values at X_cand
        """
        mu_pred, var_pred = predict(X_cand, self.model)
        lfi = D.Normal(0,1).cdf(
            (mu_pred - self.eta) / var_pred.sqrt()
        )
        if log:
            return (lfi + torch.finfo().eps).log()
        else:
            return lfi
    
    def __call__(self, X_cand, log=False):
        """
        Compute pi at given locations X_cand
        
        Args:
        - X_cand: torch.tensor, inputs X where to compute LFI
        - log: bool: return log value if true, otherwise not.
        
        Return:
        - pi: torch.tensor, the pi values at X_cand
        """
        if self.label == "ts":
            raise NotImplementedError("Not implemented yet")
        elif self.label == "lfi":
            return self.lfi(X_cand, log=log)
        else:
            raise ValueError("Label should be either 'ts' or 'lfi'.")
            
class PI_FBGP:
    def __init__(self, model):
        """
        Definition of pi (feasible resion) for fully Bayesian GP.
        
        Args:
        - model: gpytorch.models, function of GP model
        - label: string, "ts" or "lfi"
        """
        self.model = model
    
    def lfi(self, X_cand, log=False):
        """
        Compute LFI at given locations X_cand
        
        Args:
        - X_cand: torch.tensor, inputs X where to compute LFI
        - log: bool: return log value if true, otherwise not.
        
        Return:
        - lfi: torch.tensor, the LFI values at X_cand
        """
        mu_batch, var_batch = self.model.batch_predict(X_cand)
        lfi = D.Normal(0,1).cdf(
            (mu_batch - self.model.Theta_qd[:,0].unsqueeze(1))/ var_batch.sqrt()
        )
        lfi = self.model.w_qd @ lfi
        
        if log:
            return (lfi + torch.finfo().eps).log()
        else:
            return lfi
    
    def __call__(self, X_cand, log=False):
        """
        Compute pi at given locations X_cand
        
        Args:
        - X_cand: torch.tensor, inputs X where to compute LFI
        - log: bool: return log value if true, otherwise not.
        
        Return:
        - pi: torch.tensor, the pi values at X_cand
        """
        if self.label == "ts":
            raise NotImplementedError("Not implemented yet")
        elif self.label == "lfi":
            return self.lfi(X_cand, log=log)
        else:
            raise ValueError("Label should be either 'ts' or 'lfi'.")

class PI_BQ:
    def __init__(self, model):
        """
        Definition of pi (feasible resion) for Bayesian Quadrature GP model.
        For BQ modelling, see https://arxiv.org/abs/2210.17299
        
        Args:
        - model: gpytorch.models, function of GP model
        - label: string, "ts" or "lfi"
        """
        self.model = model
    
    def lfi(self, X_cand, log=False):
        """
        Compute LFI at given locations X_cand
        
        Args:
        - X_cand: torch.tensor, inputs X where to compute LFI
        - log: bool: return log value if true, otherwise not.
        
        Return:
        - lfi: torch.tensor, the LFI values at X_cand
        """
        mu_pred, var_pred = self.model.gspace_predict(X_cand)
        lfi = D.Normal(0,1).cdf(
            (mu_pred - 1) / var_pred.sqrt()
        )
        if log:
            return (lfi + torch.finfo().eps).log()
        else:
            return lfi
    
    def __call__(self, X_cand, log=False):
        """
        Compute pi at given locations X_cand
        
        Args:
        - X_cand: torch.tensor, inputs X where to compute LFI
        - log: bool: return log value if true, otherwise not.
        
        Return:
        - pi: torch.tensor, the pi values at X_cand
        """
        if self.label == "ts":
            raise NotImplementedError("Not implemented yet")
        elif self.label == "lfi":
            return self.lfi(X_cand, log=log)
        else:
            raise ValueError("Label should be either 'ts' or 'lfi'.")
    


from numpy.core.numeric import full_like, zeros_like
from scipy.stats import mvn
from numpy import array, int32,Inf
from joblib import Parallel, delayed


class Integration:
    def __init__(self):
        self.maxpts = None
        self.abseps = 1e-6
        self.releps = 1e-6
        self.n_jobs = 1

integration = Integration()

# TODO : allow more joblib control

def integrate(l,u,m,c):
    return mvn.mvnun(l, u, m, c, integration.maxpts, integration.abseps, integration.releps)

def parallel_integration(l,u,m,c):
    N = c.shape[0]
    if N==0:
        return tuple(), tuple()
    p = Parallel(n_jobs=integration.n_jobs)(
        delayed(integrate)(l[j,...], u[j,...], m[j,...], c[j,...]) for j in range(N)) 
    v, i = zip(*p)
    return v, i


def prod(tup):
    a = 1
    for i in tup:
        a *= i
    return a

def hyperrectangle_integration(mean,covariance,lower=None,upper=None,info=False):
    # parallel batch version of scipy.stats.mvn
    # no pytorch here
    # default: integration over the first orthant (for all i, Y_i<0)
    ms = mean.shape
    batch_shape = ms[:-1]
    d = ms[-1]

    N = prod(batch_shape)
    m = mean.reshape(N,d)
    l = full_like(m,-Inf) if lower is None else lower.reshape(N,d)
    u = zeros_like(m) if upper is None else upper.reshape(N,d)
    c = covariance.reshape(N,d,d)


    v, i = parallel_integration(l,u,m,c)
    values = array(v).reshape(batch_shape)
    if info :
        infos  = array(i, dtype = int32).reshape(batch_shape)
        return (values,infos)
    else:
        return values




import torch
from .._rchq import recombination
from .._sampler import MixtureSampler
from .._utils import TensorManager

class BASQ(TensorManager):
    def __init__(
        self, 
        prior, 
        model,
        sober,
        ratio_wkde=1
    ):
        """
        inference of evidence and posterior by BASQ.
        See details in https://arxiv.org/abs/2206.04734
        
        Args:
           - prior: Prior class, prior distribution
           - model: ScaleMmltGP class, Bayesian quadrature model
           - sober: Sober class, SOBER model
           - ratio_wkde: float, the proportion to sample from pi
        """
        super().__init__()
        self.prior = prior
        self.update_model(model, sober, ratio_wkde=ratio_wkde)

    def update_model(self, model, sober, ratio_wkde=1):
        """
        Update model
        
        Args:
           - model: ScaleMmltGP class, Bayesian quadrature model
           - sober: Sober class, SOBER model
           - ratio_wkde: float, the proportion to sample from pi
        """
        self.kernel = model.gspace_kernel
        self.pred_mean = model.gspace_mean_predict
        self.beta = model.beta
        self.sampler = MixtureSampler(self.prior, sober, ratio_wkde=ratio_wkde)
        
    def quadrature(self, n_quad, n_nys_quad, n_res_quad):
        """
        Kernel quadrature for the estimation of marginal likelihood
        
        Args:
           - n_quad: int, sampling size for kernel recombination
           - n_nys_quad: int, number of samples for Nyström approximation
           - n_res_quad: int, number of kernel recombination subsamples
           
        Returns:
            - ELML: float, Expected log marginal likelihood
            - AVLML: float, Variance log marginal likelihood
        """
        X_cand = self.prior.sample(n_quad)
        w_IS = self.ones(n_quad) / n_quad
        X_nys = X_cand[:n_nys_quad]
        
        idx, w = recombination(
            X_cand,
            X_nys,
            n_res_quad,
            self.kernel,
            self.device,
            self.dtype,
            init_weights=w_IS,
        )
        x = X_cand[idx]
        
        # expected log marginal likelihood
        self.EML = w @ self.pred_mean(x)
        if self.EML <= 0:
            ELML = self.beta
            self.EML = self.beta.exp()
        else:
            ELML = self.EML.log() + self.beta
        # approximated variance of log marginal likelihod
        AVLML = (w @ self.kernel(x, x) @ w).abs().log() # + 2 * model.beta
        print(f"Expected log marginal likelihood: {ELML.item():.5e}")
        print(f"Variance log marginal likelihood: {AVLML.item():.5e}")
        return ELML.item(), AVLML.item()
    
    def posterior(self, x):
        """
        Probability density function of the estimated posterior
        
        Args:
           - x: torch.tensor, the input
           
        Returns:
            - posterior_pred: torch.tensor, the expected PDF of the estimated posterior
        """
        if hasattr(self, "EML"):
            likelihood_pred = self.pred_mean(x)
            likelihood_pred[likelihood_pred < 0] = 0
            if self.EML <= 0:
                raise ValueError("Evidence is not positive.")
            else:
                posterior_pred = likelihood_pred * self.prior.pdf(x) / self.EML
            return posterior_pred
        else:
            raise ValueError("Evidence has not yet computed.")
    
    def sampling_posterior(self, n_samples, ratio_super=100):
        """
        Approximately sampling from posterior via sequential importance resampling (SIR)
        
        Args:
           - n_samples: int, number of samples to draw
           - ratio_super: float, the ratio to supersample
           
        Returns:
            - samples: torch.tensor, the samples from the estimated posterior
        """
        n_supersamples = int(ratio_super * n_samples)
        samples = self.sampler.sample(n_supersamples)
        pdf_sampler = self.sampler.pdf(samples)
        pdf_posterior = self.posterior(samples)
        weights = pdf_posterior / pdf_sampler
        weights = self.sampler.sober.cleansing_weights(weights)
        idx = self.sampler.sober.weighted_resampling(weights.detach(), n_samples)
        samples = samples[idx]
        return samples
    
    def MAP(self, n_samples):
        """
        Maximum a posteriori (MAP) estimation
        
        Args:
           - n_samples: int, number of samples to draw
           
        Returns:
            - MAP: torch.tensor, the MAP value
        """
        samples = self.sampler.sample(n_samples)
        pdf_posterior = self.posterior(samples)
        MAP = samples[pdf_posterior.argmax()]
        return MAP



from torch import tril,triu, diagonal

def make_condition(x,m,c):
    v = diagonal(c,dim1=-2,dim2=-1).unsqueeze(-1)
    c_v = c/v
    m_c = m.unsqueeze(-2)+(x-m).unsqueeze(-1)*c_v
    m_cond = remove_diagonal(m_c)

    v_columns = c_v.unsqueeze(-3)
    cun = c.unsqueeze(-1)
    c_3d = cun*v_columns
    c_c = c.unsqueeze(-2)-c_3d
    c_cond = swapcut(c_c) # d x d x d -> d x d-1 x d-1
    return m_cond, c_cond

def remove_diagonal(MU):
    return tril(MU,-1)[...,:-1]+triu(MU,1)[...,1:]

def swapcut(Gamma):
    G = remove_diagonal(Gamma)
    G = G.swapaxes(-1,-3)
    G = remove_diagonal(G)
    G = G.swapaxes(-3,-2)
    return G


if __name__ == "__main__":
    import torch
    d = 4
    m = torch.rand(d,requires_grad=True,dtype=torch.float64)
    x = torch.zeros_like(m)
    a = torch.rand(d,d,dtype=torch.float64)
    c = torch.matmul(a,a.t()).unsqueeze(0)

    m_c, c_c = make_condition(x,m,c)
    for i in range(c.shape[0]):
        print("i="+str(i))
        C = c[i,...]
        print("c_ci")
        print(c_c[i,...])
        print(C[1:3,1:3]-1/C[0,0]*C[1:3,0].unsqueeze(-1)*C[1:3,0])
        c01 = torch.tensor([C[0,1],C[2,1]])
        print(torch.tensor([[float(C[0,0]),float(C[0,2])],[float(C[2,0]),float(C[2,2])]])-1/C[1,1]*c01.unsqueeze(-1)*c01)
        print(C[0:2,0:2]-1/C[2,2]*C[0:2,2].unsqueeze(-1)*C[0:2,2])


        print(m[1:3]+1/C[0,0]*C[1:3,0]*(x[0]-m[0]))
        print(m_c)
        print(x[0:2].unsqueeze(-2).shape)
        print(m_c.shape)
        print(c_c.shape)


from itertools import zip_longest
from torch import broadcast_to, erfc, eye, tril, diagonal
from .Phi import Phi

def broadcast_shape(a,b):
    res = reversed(tuple(i if j==1 else (j if i==1 else (i if i==j else -1)) for i,j in zip_longest(reversed(a),reversed(b),fillvalue=1)))
    return list(res)

def PhiDiagonal(z):
    return (erfc(-z*0.70710678118654746171500846685)/2).prod(-1)
    #                  ^ sqrt(2)/2

def multivariate_normal_cdf(value,loc=0.0,covariance_matrix=None,diagonality_tolerance=0.0):
    """Compute orthant probabilities ``P(Z_i < value_i, i = 1,...,d)`` for a multivariate normal random vector Z.
    Closed-form backward differentiation with respect to `value`, `loc` or `covariance_matrix` is supported.

    Parameters
    ----------
    value : torch.Tensor,
        upper integration limits. It can have batch shape.
        The last dimension must be equal to d, the dimension of the
        Gaussian vector.
    loc : torch.Tensor, optional
        Mean of the Gaussian vector. Default is zeros. Can have batch
        shape. Last dimension must be equal to d, the dimension of the
        Gaussian vector. If a float is provided, the value is repeated
        for all the d components.
    covariance_matrix : torch.Tensor, optional
        Covariance matrix of the Gaussian vector.
        Can have batch shape. The two last dimensions must be equal
        to d. Identity matrix by default.
    diagonality_tolerance=0.0 : float, optional
        Avoid expensive numerical integration if the maximum of all
        off-diagonal values is below this tolerance (in absolute value),
        as the covariance is considered diagonal. If there is a batch of
        covariances (e.g. `covariance_matrix` has shape [N,d,d]), then
        the numerical integrations are avoided only if *all* covariances
        are considered diagonal. Diagonality check can be avoided with
        a negative value.
    Returns
    -------
    probability : torch.Tensor
        The probability of the event ``Y < value``. Its shape is the
        the broadcasted batch shape (just a scalar if the batchshape is []).
        Closed form derivative are implemented if `value`  `loc`,
        `covariance_matrix` require a gradient.
    Notes
    -------
    Parameters `value` and `covariance_matrix`, as 
    well as the returned probability tensor are broadcasted to their
    common batch shape. See PyTorch' `broadcasting semantics
    <https://pytorch.org/docs/stable/notes/broadcasting.html#broadcasting-semantics>`_.
    The integration is performed with Scipy's impementation of A. Genz method [1]_.
    Partial derivative are computed using closed form formula, see e.g. Marmin et al. [2]_, p 13.
    References
    ----------
    .. [1] Alan Genz and Frank Bretz, "Comparison of Methods for the Computation of Multivariate 
       t-Probabilities", Journal of Computational and Graphical Statistics 11, pp. 950-971, 2002. `Source code <http://www.math.wsu.edu/faculty/genz/software/fort77/mvtdstpack.f>`_.
    .. [2] Sébastien Marmin, Clément Chevalier and David Ginsbourger, "Differentiating the multipoint Expected Improvement for optimal batch design", International Workshop on Machine learning, Optimization and big Data, Taormina, Italy, 2015. `PDF <https://hal.archives-ouvertes.fr/hal-01133220v4/document>`_.
    Examples
    --------
    >>> import torch
    >>> from torch.autograd import grad
    >>> from mvnorm import multivariate_normal_cdf as Phi
    >>> n = 4
    >>> x = 1 + torch.randn(n)
    >>> x.requires_grad = True
    >>> # Make a positive semi-definite matrix
    >>> A = torch.randn(n,n)
    >>> C = 1/n*torch.matmul(A,A.t())
    >>> p = Phi(x,covariance_matrix=C)
    >>> p
    tensor(0.3721, grad_fn=<PhiHighDimBackward>)
    >>> grad(p,(x,))[0]
    tensor([0.0085, 0.2510, 0.1272, 0.0332])
    """
    m = loc-value # actually do P(Y-value<0)
    m_shape = m.shape
    d = m_shape[-1]
    if covariance_matrix is None:
        covariance_matrix = eye(d)
        off_diag = -0.0
    else:
        if diagonality_tolerance>=0:
            if d>=2:
                off_diag = tril(covariance_matrix.detach(),diagonal = -1).abs().max()
            else:
                off_diag = -0.0
        else: # diagonality check forbidden by user
            off_diag = diagonality_tolerance + 1
    if off_diag<=diagonality_tolerance: # assumed diagonal
        D = diagonal(covariance_matrix,dim1 = -2, dim2 = -1)
        z = -m/D.sqrt()
        return PhiDiagonal(z)
    cov_shape = covariance_matrix.shape[-2:]
    if len(cov_shape) < 2:
        raise ValueError("covariance_matrix must have at last " \
                         "two dimensions when not diagonal.")
    if cov_shape[-2] != d or cov_shape[-1] != d:
        raise ValueError("Covariance matrix must have the last two " \
                         "dimensions equal to d. Here it's "+str(list(cov_shape[-2:])))
    batch_shape = broadcast_shape(m.shape[:-1],cov_shape[:-2])
    vector_shape = batch_shape + [d]
    matrix_shape = batch_shape + [d,d]
    m_b = broadcast_to(m,vector_shape)
    c_b = broadcast_to(covariance_matrix,matrix_shape)
    return Phi(m_b,c_b)


import time
import torch
from ._utils import SafeTensorOperator
tm = SafeTensorOperator()

def recombination(
    pts_rec,         # random samples for recombination
    pts_nys,         # number of samples used for approximating kernel with Nystrom method
    num_pts,         # number of samples finally returned
    kernel,          # kernel
    device,          # device
    dtype,           # dtype
    init_weights=None,  # initial weights of the sample for recombination
    calc_obj=None,   # a function for calculating an additional objective function
):
    """
    Args:
        - pts_nys: torch.tensor, subsamples for low-rank approximation via Nyström method
        - pts_rec: torch.tensor, subsamples for empirical measure of kernel recomnbination
        - num_pts: int, number of samples finally returned. In BASQ context, this is equivalent to batch size
        - kernel: function of covariance_matrix = function(X, Y). Positive semi-definite Gram matrix (a.k.a. kernel)
        - device: torch.device, cpu or cuda
        - dtype: torch.dtype, torch.float or torch.double
        - init_weights: torch.tensor, weights for importance sampling if pts_rec is not sampled from the prior
        - calc_obj: a function that returns a Tensor of objective values for all the input points

    Returns:
        - x: torch.tensor, the sparcified samples from pts_rec. The number of samples are determined by self.batch_size
        - w: torch.tensor, the positive weights for kernel quadrature as discretised summation.
    """
    return rc_kernel_svd(pts_rec, pts_nys, num_pts, kernel, tm, mu=init_weights, calc_obj=calc_obj)


def ker_svd_sparsify(pt, s, kernel):
    mat = kernel(pt, pt)
    mat = tm.make_cov_psd(mat)
    _U, S, _ = torch.svd_lowrank(mat, q=s)
    U = -1 * _U.T  # Hermitian
    return S, U


def rc_kernel_svd(samp, pt, s, kernel, tm, mu=None, calc_obj=None):
    # Nystrom method
    _, U = ker_svd_sparsify(pt, s - 1, kernel)
    w_star, idx_star = Mod_Tchernychova_Lyons(
        samp, U, pt, kernel, tm, mu=mu, calc_obj=calc_obj
    )
    return idx_star, w_star


def Mod_Tchernychova_Lyons(samp, U_svd, pt_nys, kernel, tm, mu=None, calc_obj=None, DEBUG=False):
    """
    This function is a modified Tcherynychova_Lyons from
    https://github.com/FraCose/Recombination_Random_Algos/blob/master/recombination.py
    """
    N = len(samp)
    n, length = U_svd.shape
    number_of_sets = 2 * (n + 1)

    if mu is None:
        mu = tm.ones(N) / N

    idx_story = tm.arange(N)
    idx_story = idx_story[mu != 0]
    remaining_points = len(idx_story)

    use_obj = False if calc_obj is None else True
    if use_obj:
        obj = -1 * calc_obj(samp)

    while True:
        if remaining_points <= n + 1:
            idx_star = tm.arange(len(mu))[mu > 0]
            w_star = mu[idx_star]
            return w_star, idx_star

        elif n + 1 < remaining_points <= number_of_sets:
            X_mat = U_svd @ kernel(pt_nys, samp[idx_story])
            if use_obj:
                X_mat = torch.cat((X_mat, torch.reshape(
                    obj[idx_story], (1, -1))), 0)
                X_mat_raw = torch.clone(X_mat[:-1])
            
            w_star, idx_star, x_star, _, ERR, _, _ = Tchernychova_Lyons_CAR(
                X_mat.T, torch.clone(mu[idx_story]), tm, DEBUG)

            if use_obj:
                Xp = X_mat_raw[:, idx_star]
                obj_p = obj[idx_star]
                Xp = torch.cat((Xp, tm.ones(1, len(idx_star))), 0)
                _, _, w_null = torch.linalg.svd(Xp)
                w_null = w_null[-1]
                if torch.dot(obj_p, w_null) < 0:
                    w_null = -w_null

                lm = len(w_star)
                plis = w_null > 0
                alpha = tm.zeros(lm)
                alpha[plis] = w_star[plis] / w_null[plis]
                idx_sp = tm.arange(lm)[plis]
                idx_sp = idx_sp[torch.argmin(alpha[plis])]
                w_star = w_star-alpha[idx_sp]*w_null
                w_star[idx_sp] = 0.

                idx_star = idx_star[w_star > 0]
                w_star = w_star[w_star > 0]

            idx_story = idx_story[idx_star]
            mu[:] = 0.
            mu[idx_story] = w_star
            idx_star = idx_story
            w_star = mu[mu > 0]

            return w_star, idx_star

        number_of_el = int(remaining_points / number_of_sets)

        idx = idx_story[:number_of_el *
                        number_of_sets].reshape(number_of_el, -1)
        
        # compute X_for_nys and X_for_obj
        N_approx = number_of_sets * number_of_el
        _idx_tmp = idx_story[:N_approx].reshape(number_of_el, number_of_sets)
        K = kernel(pt_nys, samp[_idx_tmp]) * mu[_idx_tmp].unsqueeze(1)
        X_for_nys = tm.zeros(length, number_of_sets)
        X_for_nys += K.sum(axis=0)

        N_rest = len(idx_story) - N_approx
        if N_rest > 0:
            idx_rest = idx_story[N_approx : N_approx + N_rest]
            K = kernel(pt_nys, samp[idx_rest]) * mu[idx_rest].unsqueeze(0)
            K_pad = torch.cat(
                (K, tm.zeros(length, number_of_sets - N_rest)),
                dim=1,
            )
            X_for_nys += K_pad

        if use_obj:
            X_for_obj = tm.zeros(1, number_of_sets)
            X_for_obj += (
                obj[_idx_tmp].unsqueeze(1) * mu[_idx_tmp].unsqueeze(1)
            ).sum(axis=0).reshape(-1,1)
            if N_rest > 0:
                mat_obj = (obj[idx_rest].unsqueeze(0) * mu[idx_rest].unsqueeze(0)).reshape(-1,1)
                mat_obj_pad = torch.cat((mat_obj, tm.zeros(number_of_sets - N_rest, 1)), dim=0)
                X_for_obj += mat_obj_pad        
        
        X_tmp_tr = U_svd @ X_for_nys
        if use_obj:
            X_tmp_tr = torch.cat((X_tmp_tr, X_for_obj), 0)
        X_tmp = X_tmp_tr.T
        tot_weights = torch.sum(mu[idx], 0)
        idx_last_part = idx_story[number_of_el * number_of_sets:]

        if len(idx_last_part):
            X_mat = U_svd @ kernel(pt_nys, samp[idx_last_part])
            if use_obj:
                X_mat = torch.cat((X_mat, torch.reshape(
                    obj[idx_last_part], (1, -1))), 0)
            X_tmp[-1] += torch.multiply(
                X_mat.T,
                mu[idx_last_part].unsqueeze(1)
            ).sum(axis=0)
            tot_weights[-1] += torch.sum(mu[idx_last_part], 0)
        
        X_tmp = torch.divide(X_tmp, tot_weights.unsqueeze(0).T)

        # sparsify for the case use_obj is True
        if use_obj:
            X_tmp_raw = torch.clone(X_tmp[:, :n])
            obj_raw = X_tmp[:, -1:].reshape(-1)
        
        w_star, idx_star, _, _, ERR, _, _ = Tchernychova_Lyons_CAR(
            X_tmp, torch.clone(tot_weights), tm
        )

        if use_obj:
            Xp = X_tmp_raw[idx_star].T
            obj_p = obj_raw[idx_star]
            Xp = torch.cat((Xp, tm.ones(1, len(idx_star))), 0)
            _, _, w_null = torch.linalg.svd(Xp)
            w_null = w_null[-1]
            if torch.dot(obj_p, w_null) < 0:
                w_null = -w_null

            lm = len(w_star)
            plis = w_null > 0
            alpha = tm.zeros(lm)
            alpha[plis] = w_star[plis] / w_null[plis]
            idx_sp = tm.arange(lm)[plis]
            idx_sp = idx_sp[torch.argmin(alpha[plis])]
            w_star = w_star-alpha[idx_sp]*w_null
            w_star[idx_sp] = 0.

            idx_star = idx_star[w_star > 0]
            w_star = w_star[w_star > 0]

        idx_tomaintain = idx[:, idx_star].reshape(-1)
        idx_tocancel = tm.ones(idx.shape[1]).to(torch.bool)
        idx_tocancel[idx_star] = 0
        idx_tocancel = idx[:, idx_tocancel].reshape(-1)

        mu[idx_tocancel] = 0.
        mu_tmp = torch.multiply(mu[idx[:, idx_star]], w_star)
        mu_tmp = torch.divide(mu_tmp, tot_weights[idx_star])
        mu[idx_tomaintain] = mu_tmp.reshape(-1)

        idx_tmp = idx_star == number_of_sets - 1
        idx_tmp = tm.arange(len(idx_tmp))[idx_tmp != 0]
        # if idx_star contains the last barycenter, whose set could have more points
        if len(idx_tmp) > 0:
            mu_tmp = torch.multiply(mu[idx_last_part], w_star[idx_tmp])
            mu_tmp = torch.divide(mu_tmp, tot_weights[idx_star[idx_tmp]])
            mu[idx_last_part] = mu_tmp
            idx_tomaintain = torch.cat([idx_tomaintain, idx_last_part])
        else:
            idx_tocancel = torch.cat([idx_tocancel, idx_last_part])
            mu[idx_last_part] = 0.

        idx_story = torch.clone(idx_tomaintain)
        remaining_points = len(idx_story)


def Tchernychova_Lyons_CAR(X, mu, tm, DEBUG=False):
    """
    This functions reduce X from N points to n+1.
    This is taken from https://github.com/FraCose/Recombination_Random_Algos/blob/master/recombination.py
    """
    X = torch.cat([tm.ones(X.size(0)).unsqueeze(0).T, X], dim=1)
    N, n = X.shape
    U, Sigma, V = torch.linalg.svd(X.T)
    U = torch.cat([U, tm.zeros(n, N - n)], dim=1)
    Sigma = torch.cat([Sigma, tm.zeros(N - n)])
    Phi = V[-(N - n):, :].T
    cancelled = tm.null()

    for _ in range(N - n):
        lm = len(mu)
        plis = Phi[:, 0] > 0
        # Added 7th Aug, 2023.
        if plis.sum() == 0:
            break
        
        alpha = tm.zeros(lm)
        alpha[plis] = mu[plis] / Phi[plis, 0]
        idx = tm.arange(lm)[plis]
        idx = idx[torch.argmin(alpha[plis])]

        if len(cancelled) == 0:
            cancelled = idx.unsqueeze(0)
        else:
            cancelled = torch.cat([cancelled, idx.unsqueeze(0)])
        mu[:] = mu - alpha[idx] * Phi[:, 0]
        mu[idx] = 0.

        if DEBUG and (not torch.allclose(torch.sum(mu), 1.)):
            # print("ERROR")
            print("sum ", torch.sum(mu))

        Phi_tmp = Phi[:, 0]
        Phi = Phi[:, 1:]
        Phi = Phi - torch.matmul(
            Phi[idx].unsqueeze(1),
            Phi_tmp.unsqueeze(1).T,
        ).T / Phi_tmp[idx]
        Phi[idx, :] = 0.

    w_star = mu[mu > 0]
    idx_star = tm.arange(N)[mu > 0]
    return w_star, idx_star, torch.nan, torch.nan, 0., torch.nan, torch.nan



import torch

_device = (
    torch.device('cuda')
    if torch.cuda.is_available()
    else torch.device('cpu')
)
_dtype = torch.double


def setting_parameters(device=None, dtype=None):
    """
    Return:
       - device: torch.device, cpu or cuda
       - dtype: torch.dtype, torch.float or torch.double
    """
    global _device, _dtype
    if device:
        _device = device
    if dtype:
        _dtype = dtype
    return _device, _dtype



import torch
import warnings

class WeightsStabiliser:
    def __init__(
        self,
        eps=torch.finfo().eps,
        thresh=5,
    ):
        """
        A class of functions that stabilise the weight-related computations
        
        Args:
        - eps_weights: float, the machine epsilon (the smallest number of floating point).
                       Default: torch.finfo().eps
        - thresh: int, the number of non-zero weights which regrads anomalies.
        """
        self.eps_weights = eps
        self.thresh = thresh
        
    def cleansing_weights(self, weights):
        """
        Remove anomalies from the computed weights
        
        Args:
        - weights: torch.tensor, weights
        
        Return:
        - weights: torch.tensor, the cleaned weights
        """
        weights[weights < self.eps_weights] = 0
        weights[weights.isinf()] = self.eps_weights
        weights[weights.isnan()] = self.eps_weights
        if not weights.sum() == 0:
            weights /= weights.sum()
        else:
            weights = torch.ones_like(weights)/len(weights)
        return weights.detach()
    
    def check_weights(self, weights):
        """
        Check weights anomalies
        
        Args:
        - weights: torch.tensor, weights
        
        Return:
        - weights: torch.tensor, the cleaned weights
        """
        if weights.sum() == 0:
            return False
        elif len(weights.unique()) < self.thresh:
            return False
        else:
            return True
        
    def weighted_resampling(self, weights, n_nys):
        """
        Weighted resampling.
        len(weights) > n_nys should be satisfied.
        
        Args:
        - weights: torch.tensor, weights
        - n_nys: int, the number of resamples
        
        Return:
        - idx_nys: torch.tensor, the indices where the resamples locate.
        """
        n_positive_weights = (weights > 0).sum()
        if n_positive_weights > n_nys:
            idx_nys = torch.multinomial(weights, n_nys)
        else:
            idx_positive = torch.arange(len(weights))[weights > 0]
            idx_rand = torch.randperm(len(weights))[:int(n_nys - n_positive_weights)]
            idx_nys = torch.cat([idx_positive, idx_rand])
            warnings.warn("Non-zero weights are fewer than n_Nys: "+str(idx_nys.sum()))
        return idx_nys
    
    def deweighted_resampling(self, weights, n_samples):
        """
        Uniform resampling from weighted samples
        
        Args:
        - weights: torch.tensor, the unnormalised weights
        - n_samples, int, the number of uniform samples
        
        Return:
        - indice: torch.tensor, the indices of the selected uniform samples
        """
        weights_inv = (1 / weights)
        weights_inv = self.cleansing_weights(weights_inv)
        indice = self.weighted_resampling(weights_inv, n_samples)
        return indice
    
    def kmeans_resampling(self, X, n_clusters=100):
        _, X_sparse = KMeans(X, n_clusters)
        return X_sparse


def KMeans(x, K=10, Niter=10):
    """Implements Lloyd's algorithm for the Euclidean metric."""
    N, D = x.shape  # Number of samples, dimension of the ambient space
    c = x[:K, :].clone()  # Simplistic initialization for the centroids

    x_i = x.view(N, 1, D)  # (N, 1, D) samples
    c_j = c.view(1, K, D)  # (1, K, D) centroids

    # K-means loop:
    # - x  is the (N, D) point cloud,
    # - cl is the (N,) vector of class labels
    # - c  is the (K, D) cloud of cluster centroids
    for i in range(Niter):
        # E step: assign points to the closest cluster -------------------------
        D_ij = ((x_i - c_j) ** 2).sum(-1)  # (N, K) symbolic squared distances
        cl = D_ij.argmin(dim=1).long().view(-1)  # Points -> Nearest cluster

        # M step: update the centroids to the normalized cluster average: ------
        # Compute the sum of points per cluster:
        c.zero_()
        c.scatter_add_(0, cl[:, None].repeat(1, D), x)

        # Divide by the number of points per cluster:
        Ncl = torch.bincount(cl, minlength=K).type_as(c).view(K, 1)
        c /= Ncl  # in-place division to compute the average

    return cl, c


from ._gp import predictive_covariance, predict_mean
import torch.distributions as D

class Kernel:
    def __init__(self, model, mode="predictive_covariance"):
        """
        Definition of kernel for recombination.
        
        Args:
        - model: gpytorch.models, function of GP model
        - mode: string, select from ["predictive_covariance", "weighted_predictive_covariance", "kernel"]
        """
        self.model = model
        self.mode = mode
    
    def __call__(self, x, y):
        """
        Compute the Gram matrix of posterior predictive covariance
        
        Return:
        - CLy: torch.tensor, the Gram matrix with the posterior predictive covariance
        """
        if self.mode == "predictive_covariance":
            return predictive_covariance(x, y, self.model)
        elif self.mode == "weighted_predictive_covariance":
            return self.weighted_covariance(x, y)
        elif self.mode == "kernel":
            return self.model.covar_module.forward(x, y)
        else:
            raise ValueError('mode should be from ["predictive_covariance", "weighted_predictive_covariance", "kernel"]')
            
    
    def weighted_covariance(self, x, y):
        """
        Compute the mean weighted Gram matrix
        
        Return:
        - CLy: torch.tensor, the Gram matrix with the posterior predictive covariance
        """
        mu_x = predict_mean(x, self.model)
        mu_y = predict_mean(y, self.model)
        cov_xy = predictive_covariance(x, y, self.model)
        if len(mu_x.shape) == 1 and len(mu_y.shape) == 1:
            CLy = mu_x.unsqueeze(1) * cov_xy * mu_y.unsqueeze(0)
        else:
            CLy = mu_x.unsqueeze(1) * cov_xy * mu_y.unsqueeze(1)
        return CLy



from numpy import zeros_like as np_zeros_like
from torch import tensor, diagonal, zeros, tril_indices, tril, triu, diag_embed, erfc, from_numpy
from torch.autograd import Function

from .integration import hyperrectangle_integration
from .conditioning import make_condition


def phi(z,v):
    return 0.39894228040143270286321808271/v.sqrt()*(-z**2/(2*v)).exp()
    #              ^ oneOverSqrt2pi     

def Phi1D(x,m,c):
    z = (x-m)/c.squeeze(-1).sqrt()
    return (erfc(-z*0.70710678118654746171500846685)/2).squeeze(-1)
    #                  ^ sqrt(2)/2

def phi2_sub(z,C): # compute pairs of bivariate densities and organize them in a matrix
    V = diagonal(C,dim1=-2,dim2=-1)
    a,c = V.unsqueeze(-1),V.unsqueeze(-2)
    det = a*c-C**2
    x1 = z.unsqueeze(-1)
    x2 = z.unsqueeze(-2)
    exponent = -0.5/det*(c*x1**2+a*x2**2-2*C*x1*x2)
    res = 0.15915494309189534560822210096/det.sqrt()*(exponent).exp()
    #           ^ oneOver2pi     
    return tril(res,-1) + triu(res,1)


def to_torch(x):
    if len(x.shape) == 0:
        return tensor(float(x))
    else:
        return from_numpy(x)


class PhiHighDim(Function):

    @staticmethod
    def forward(ctx, m, c):
        m_np = m.numpy()
        c_np = .5*(c+c.transpose(-1,-2)).numpy()
        ctx.save_for_backward(m,c)
        res_np = hyperrectangle_integration(m_np,c_np)
        return to_torch(res_np)

    @staticmethod
    def backward(ctx,grad_output):
        if grad_output is None:
            return None, None
        res_m = res_c = None
        need_m, need_c = ctx.needs_input_grad[0:2]
        if need_c or need_m:
            m,c = ctx.saved_tensors
            m_cond,c_cond = make_condition(0,m,c)
            v = diagonal(c,dim1=-2,dim2=-1)
            p = phi(m,v)
            P = Phi(m_cond,c_cond)
            grad_m = -P*p
            grad_output_u1 = grad_output.unsqueeze(-1)
            res_m = grad_output_u1*grad_m
            if need_c:
                d = c.shape[-1] # d==1 should never happen here
                if d==2:
                    P2 = 1
                else:
                    trilind = tril_indices(d,d-1,offset=-1)
                    m_cond2,c_cond2 = make_condition(0,m_cond,c_cond)
                    Q_l = Phi(m_cond2[...,trilind[0],trilind[1],:],c_cond2[...,trilind[0],trilind[1],:,:])
                    P2 = zeros(*Q_l.shape[:-1],d,d,dtype=Q_l.dtype)
                    P2[...,trilind[0],trilind[1]] = Q_l
                    P2[...,trilind[1],trilind[0]] = Q_l
                p2 = phi2_sub(m,c)
                hess = p2*P2
                D = -(m*grad_m+(hess*c).sum(-1))/v
                grad_c = .5*(hess + diag_embed(D))
                res_c = grad_output_u1.unsqueeze(-1)*grad_c
        return res_m, res_c

Phinception = PhiHighDim.apply

def Phi(m,c):
    d = c.shape[-1]
    if d == 1:
        return Phi1D(0,m,c)
    else:
        return Phinception(m,c)

if __name__ == "__main__":
    import torch                                                                                          
                                                                                                      
    def jacobian(y, x, create_graph=False):                                                               
        jac = []                                                                                          
        flat_y = (y*1).reshape(-1) # does someone know why it doesn't work without this one?
        grad_y = torch.zeros_like(flat_y)                                                                 
        for i in range(len(flat_y)):                                                                      
            grad_y[i] = 1.                                                                                
            grad_x, = torch.autograd.grad(flat_y, x, grad_y, retain_graph=True, create_graph=create_graph)
            jac.append(grad_x.reshape(x.shape))                                                           
            grad_y[i] = 0.                                                                                
        return torch.stack(jac).reshape(y.shape + x.shape)                                                
                                                                                                        
    def hessian(y, x):                                                                                    
        return jacobian(jacobian(y, x, create_graph=True), x)                                             
                                                                                                        
    def f(x):                                                                                             
        return x**2 +x[0]*x[1]*1
                                                                                                        
    x = torch.ones(4, requires_grad=True)                                                                 

    d = 20

    batch_shape = []

    m = torch.rand( d,requires_grad=True)#,dtype=torch.float64)
    a = torch.rand(*batch_shape,d,d)#,dtype=torch.float64)
    c = torch.matmul(a,a.transpose(-1,-2))
    
  
    P = (Phi(m,c))
    print("P =",end=" ")
    print(P)
    import time
    t1 = time.time()
    print(.5*hessian(P,m))
    t2 = time.time()
    print(t2-t1)
    c.requires_grad = True
    P = (Phi(m,c))
    t1 = time.time()
    print(grad(P,[c])[0])
    t2 = time.time()
    print(t2-t1)



import torch
from torch.distributions.normal import Normal


class FBGPAcquisitionFunction:
    def __init__(self, model, label="MES"):
        """
        Acquisition functions defined by fully Bayesian Gaussian process (FBGP) model
        
        Args:
        - model: FullyBayesianGP class, a function of FBGP model.
        - label: string, select from ["EI", "UCB", "MES", "BQBC", "QBMGP"],
                 EI: expected improvement
                 UCB: upper confidence bound
                 MES: max-value entropy search
                 BQBC: Bayesian query-by-committee
                 QBMGP: Query by a mixture of Gaussian processes
        """
        self.model = model
        self.label = label
    
    def EI(self, mu_batch, var_batch):
        """
        Expected improvement acquisition function
        
        Args:
        - mu_batch: torch.tensor, a batch of predictive mean from FBGP model
        - var_batch: torch.tensor, a batch of predictive variance from FBGP model
        
        Return:
        - af: torch.tensor, the result of acquisition function computation
        """
        integrand = (mu_batch - self.model.Theta_qd[:,0].unsqueeze(1)) / var_batch.sqrt()
        Phi = Normal(0,1).cdf(integrand)
        phi = Normal(0,1).log_prob(integrand).exp()
        first = (mu_batch - self.model.Theta_qd[:,0].unsqueeze(1)) * Phi
        second = var_batch.sqrt() * phi
        return self.model.w_qd @ (first + second)

    def UCB(self, mu_batch, var_batch):
        """
        Upper confidence bound acquisition function
        
        Args:
        - mu_batch: torch.tensor, a batch of predictive mean from FBGP model
        - var_batch: torch.tensor, a batch of predictive variance from FBGP model
        
        Return:
        - af: torch.tensor, the result of acquisition function computation
        """
        Ey = self.model.w_qd @ mu_batch
        Vy = self.model.w_qd @ (var_batch + mu_batch.pow(2)) - Ey**2
        return Ey + Vy.sqrt()

    def FITBO(self, mu_batch, var_batch):
        """
        Max-value entropy search acquisition function.
        Approximated by FITBO formulation (https://arxiv.org/abs/1711.00673)
        
        Args:
        - mu_batch: torch.tensor, a batch of predictive mean from FBGP model
        - var_batch: torch.tensor, a batch of predictive variance from FBGP model
        
        Return:
        - af: torch.tensor, the result of acquisition function computation
        """
        Ey = self.model.w_qd @ mu_batch
        Vary = self.model.w_qd @ (var_batch + mu_batch.pow(2)) - Ey**2
        H1 = 0.5 * (2 * torch.pi * torch.e * (Vary + self.model.w_qd @ self.model.Theta_qd[:,1])).log()
        H2 = 0.5 * self.model.w_qd @ (2 * torch.pi * torch.e * (var_batch + self.model.Theta_qd[:,1].unsqueeze(1))).log()
        return H1 - H2

    def BQBC(self, mu_batch):
        """
        Bayesian query-by-committee acquisition function.
        (https://arxiv.org/abs/2205.10186)
        
        Args:
        - mu_batch: torch.tensor, a batch of predictive mean from FBGP model
        - var_batch: torch.tensor, a batch of predictive variance from FBGP model
        
        Return:
        - af: torch.tensor, the result of acquisition function computation
        """
        Ey = self.model.w_qd @ mu_batch
        return self.model.w_qd @ (mu_batch - Ey)

    def QBMGP(self, mu_batch, var_batch):
        """
        Query by a mixture of Gaussian processes acquisition function.
        (https://arxiv.org/abs/2205.10186)
        
        Args:
        - mu_batch: torch.tensor, a batch of predictive mean from FBGP model
        - var_batch: torch.tensor, a batch of predictive variance from FBGP model
        
        Return:
        - af: torch.tensor, the result of acquisition function computation
        """
        Ey = self.model.w_qd @ mu_batch
        Vy = self.model.w_qd @ (var_batch + mu_batch.pow(2)) - Ey**2
        return Vy + self.BQBC(mu_batch)
    
    def __call__(self, x):
        mu_batch, var_batch = self.model.batch_predict(x)
        if self.label == "EI":
            return self.EI(mu_batch, var_batch)
        elif self.label == "UCB":
            return self.UCB(mu_batch, var_batch)
        elif self.label == "MES":
            return self.FITBO(mu_batch, var_batch)
        elif self.label == "BQBC":
            return self.BQBC(mu_batch)
        elif self.label == "QBMGP":
            return self.QBMGP(mu_batch, var_batch)
        else:
            raise ValueError("Acquisition function type should be from ['EI', 'UCB','MES', 'BQBC', 'QBMGP']")
        
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    
    



import copy
import torch
from .._gp import set_gp, update_gp, predict, predictive_covariance
from .._utils import Utils


class FitboGP:
    def __init__(
        self,
        Xobs,
        Yobs,
        gp_kernel,
        device,
        label="wsabim",
        alpha_factor=1,
        lik=1e-10,
        training_iter=10000,
        thresh=0.01,
        lr=0.1,
        rng=10,
        train_lik=False,
        optimiser="BoTorch",
    ):
        """
        FITBO (WSABI BQ) modelling
        FITBO -> https://arxiv.org/abs/1711.00673
        WSABI -> https://arxiv.org/abs/1411.0439
        WsabiGP class summarises the functions of training, updating the warped GP model.
        This also provides the prediction and kernel of WSABI GP.
        The modelling of WSABI-L and WSABI-M can be easily switched by changing "label".

        Args:
           - Xobs: torch.tensor, X samples, X belongs to prior measure.
           - Yobs: torch.tensor, Y observations, Y = true_likelihood(X).
           - gp_kernel: gpytorch.kernels, GP kernel function
           - device: torch.device, device, cpu or cuda
           - label: string, the wsabi type, ["wsabil", "wsabim"]
           - lik: float, the initial value of GP likelihood noise variance
           - train_iter: int, the maximum iteration for GP hyperparameter training.
           - thresh: float, the threshold as a stopping criterion of GP hyperparameter training.
           - lr: float, the learning rate of Adam optimiser
           - rng: int, tne range coefficient of GP likelihood noise variance
           - train_like: bool, flag whether or not to update GP likelihood noise variance
           - optimiser: string, select the optimiser ["L-BFGS-B", "Adam"]
        """
        self.gp_kernel = gp_kernel
        self.device = device
        self.alpha_factor = alpha_factor
        self.lik = lik
        self.training_iter = training_iter
        self.thresh = thresh
        self.lr = lr
        self.rng = rng
        self.train_lik = train_lik
        self.optimiser = optimiser

        self.jitter = 0  # 1e-6
        self.Y_unwarp = copy.deepcopy(Yobs)
        self.utils = Utils(device)

        self.model = update_gp(
            Xobs,
            self.process_y_warping(Yobs),
            gp_kernel,
            self.device,
            lik=self.lik,
            training_iter=self.training_iter,
            thresh=self.thresh,
            lr=self.lr,
            rng=self.rng,
            train_lik=self.train_lik,
            optimiser=self.optimiser,
        )
        self.setting(label)

    def setting(self, label):
        """
        Args:
           - label: string, the wsabi type, ["wsabil", "wsabim"]
        """
        if label == "wsabil":
            self.kernel = self.wsabil_kernel
            self.predict = self.wsabil_predict
            self.predict_mean = self.wsabil_mean_predict
        elif label == "wsabim":
            self.kernel = self.wsabim_kernel
            self.predict = self.wsabim_predict
            self.predict_mean = self.wsabim_mean_predict

    def warp_y(self, y):
        """
        Args:
           - y: torch.tensor, observations

        Returns:
           - y: torch.tensor, warped observations
        """
        return self.alpha.sign() * torch.sqrt(2 * (self.alpha - y))

    def unwarp_y(self, y):
        """
        Args:
           - y: torch.tensor, warped observations

        Returns:
           - y: torch.tensor, unwarped observations
        """
        return self.alpha - 0.5 * (y ** 2)

    def process_y_warping(self, y):
        """
        Args:
           - y: torch.tensor, observations

        Returns:
           - y: torch.tensor, warped observations that contains no anomalies and the updated alpha hyperparameter.
        """
        y = self.utils.remove_anomalies(y)
        #y -= y.min() 
        #y = (y - y.mean()) / y.std()
        self.alpha = self.alpha_factor * torch.max(y)
        y = self.warp_y(y)
        return y

    def cat_observations(self, X, Y):
        """
        Args:
           - X: torch.tensor, X samples to be added to the existing data Xobs
           - Y: torch.tensor, unwarped Y observations to be added to the existing data Yobs

        Returns:
           - Xall: torch.tensor, X samples that contains all samples
           - Yall: torch.tensor, warped Y observations that contains all observations
        """
        Xobs = self.model.train_inputs[0]
        Yobs = copy.deepcopy(self.Y_unwarp)
        if len(self.model.train_targets.shape) == 0:
            Yobs = Yobs.unsqueeze(0)
        Xall = torch.cat([Xobs, X])
        _Yall = torch.cat([Yobs, Y])
        self.Y_unwarp = copy.deepcopy(_Yall)
        Yall = self.process_y_warping(_Yall)
        return Xall, Yall

    def update_wsabi_gp(self, X, Y):
        """
        Args:
           - X: torch.tensor, X samples to be added to the existing data Xobs
           - Y: torch.tensor, unwarped Y observations to be added to the existing data Yobs
        """
        X_warp, Y_warp = self.cat_observations(X, Y)
        self.model = update_gp(
            X_warp,
            Y_warp,
            self.gp_kernel,
            self.device,
            lik=self.lik,
            training_iter=self.training_iter,
            thresh=self.thresh,
            lr=self.lr,
            rng=self.rng,
            train_lik=self.train_lik,
            optimiser=self.optimiser,
        )
        """
        self.model = set_gp(
            X_warp,
            Y_warp,
            self.gp_kernel,
            self.device,
            lik=self.lik,
            rng=self.rng,
            train_lik=self.train_lik,
        )
        """

    def retrain_gp(self):
        X_warp = self.model.train_inputs[0]
        Y_warp = self.process_y_warping(copy.deepcopy(self.Y_unwarp))
        self.model = update_gp(
            X_warp,
            Y_warp,
            self.gp_kernel,
            self.device,
            lik=self.lik,
            training_iter=self.training_iter,
            thresh=self.thresh,
            lr=self.lr,
            rng=self.rng,
            train_lik=self.train_lik,
            optimiser=self.optimiser,
        )

    def memorise_parameters(self):
        self.likelihood_memory = copy.deepcopy(torch.tensor(self.model.likelihood.noise.item()))
        self.outputsacle_memory = copy.deepcopy(torch.tensor(self.model.covar_module.outputscale.item()))
        self.lengthscale_memory = copy.deepcopy(torch.tensor(self.model.covar_module.base_kernel.lengthscale.item()))

    def remind_parameters(self):
        hypers = {
            'likelihood.noise_covar.noise': self.likelihood_memory,
            'covar_module.outputscale': self.outputsacle_memory,
            'covar_module.base_kernel.lengthscale': self.lengthscale_memory,
        }
        self.model.initialize(**hypers)

    def predictive_kernel(self, x, y):
        """
        Args:
           - x: torch.tensor, x locations to be predicted
           - y: torch.tensor, y locations to be predicted

        Args:
           - CLy: torch.tensor, the positive semi-definite Gram matrix of predictive variance
        """
        return predictive_covariance(x, y, self.model)

    def wsabil_kernel(self, x, y):
        """
        Args:
           - x: torch.tensor, x locations to be predicted
           - y: torch.tensor, y locations to be predicted

        Returns:
           - CLy: torch.tensor, the positive semi-definite Gram matrix of WSABI-L variance
        """
        mu_x, _ = predict(x, self.model)
        mu_y, _ = predict(y, self.model)
        cov_xy = predictive_covariance(x, y, self.model)
        CLy = mu_x.unsqueeze(1) * cov_xy * mu_y.unsqueeze(0)

        d = min(len(x), len(y))
        CLy[range(d), range(d)] = CLy[range(d), range(d)] + self.jitter
        return CLy

    def wsabim_kernel(self, x, y):
        """
        Args:
           - x: torch.tensor, x locations to be predicted
           - y: torch.tensor, y locations to be predicted

        Returns:
           - CLy: torch.tensor, the positive semi-definite Gram matrix of WSABI-M variance
        """
        mu_x, _ = predict(x, self.model)
        mu_y, _ = predict(y, self.model)
        cov_xy = predictive_covariance(x, y, self.model)
        CLy = mu_x.unsqueeze(1) * cov_xy * mu_y.unsqueeze(0) + 0.5 * (cov_xy ** 2)

        d = min(len(x), len(y))
        CLy[range(d), range(d)] = CLy[range(d), range(d)] + self.jitter
        return CLy

    def wsabil_predict(self, x):
        """
        Args:
           - x: torch.tensor, x locations to be predicted

        Returns:
           - mu: torch.tensor, unwarped predictive mean at given locations x.
           - var: torch.tensor, unwarped predictive variance at given locations x.
        """
        mu_warp, var_warp = predict(x, self.model)
        mu = self.alpha - 0.5 * mu_warp**2
        var = mu_warp * var_warp * mu_warp
        return mu, var

    def wsabim_predict(self, x):
        """
        Args:
           - x: torch.tensor, x locations to be predicted

        Returns:
           - mu: torch.tensor, unwarped predictive mean at given locations x.
           - var: torch.tensor, unwarped predictive variance at given locations x.
        """
        mu_warp, var_warp = predict(x, self.model)
        mu = self.alpha - 0.5 * (mu_warp**2 + var_warp)
        var = mu_warp * var_warp * mu_warp + 0.5 * (var_warp ** 2)
        return mu, var

    def wsabil_mean_predict(self, x):
        """
        Args:
           - x: torch.tensor, x locations to be predicted

        Returns:
           - mu: torch.tensor, unwarped predictive mean at given locations x.
        """
        mu_warp, _ = predict(x, self.model)
        mu = self.alpha - 0.5 * mu_warp**2
        return mu

    def wsabim_mean_predict(self, x):
        """
        Args:
           - x: torch.tensor, x locations to be predicted

        Returns:
           - mu: torch.tensor, unwarped predictive mean at given locations x.
        """
        mu_warp, var_warp = predict(x, self.model)
        mu = self.alpha - 0.5 * (mu_warp**2 + var_warp)
        return mu



import copy
import torch
import warnings
import matplotlib.pyplot as plt
from torch.distributions.multivariate_normal import MultivariateNormal
from ._utils import SafeTensorOperator
from ._weights import WeightsStabiliser
from ._prior import BasePrior
from .mvnorm import multivariate_normal_cdf as Phi

class WeightedKernelDensityEstimation(
    WeightsStabiliser, 
    SafeTensorOperator, 
    BasePrior,
):
    def __init__(
        self,
        X,
        W,
        n_dims,
        bounds=None,
        n_kde=4096,
        bw_method='scott',
        compute_cdf=False,
    ):
        """
        Class of weighted kernel density estimation with Gaussian kernel.
        
        Args:
        - X: torch.tensor, the observed data X
        - W: torch.tensor, the weights over X
        - n_dims: int, the number of dimensions
        - bounds: torch.tensor, the lower and upper bounds for each dimension. If none, the bounds are ignored.
        - n_kde: int, the number of Gaussians for KDE.
        - bw_method: string, 'scott' or 'silverman'
        - compute_cdf: bool, compute normalised PDF of truncated multivariate normal distributions if true, otherwise not.
                             The default is false as it produces significant overhead.
        """
        WeightsStabiliser.__init__(self, eps=0, thresh=n_kde) # inherit WeightsStabiliser class
        SafeTensorOperator.__init__(self) # inherit SafeTensorOperator class
        BasePrior.__init__(self) # inherit BasePrior class
        self.n_dims = n_dims
        self.bounds = bounds
        self.n_kde_init = min([n_kde, len(X)])
        self.bw_method = bw_method
        self.compute_cdf = compute_cdf
        self.type = "continuous"
        self.initialisation(X, W)
        
    def initialise_n_kde(self):
        self.n_kde = copy.deepcopy(self.n_kde_init)
    
    def initialisation(self, X, Y):
        """
        Estimate hyperparameters and sparse Gaussians
        
        Args:
        - X: torch.tensor, the observed data X
        - Y: torch.tensor, the unnormalised weights
        """
        self.initialise_n_kde()
        if self.check_weights(Y):
            idx_accept = self.deweighted_resampling(Y, self.n_kde)
        else:
            idx_accept = self.arange(Y.size(0))[self.cleansing_weights(Y) > 0]
            self.n_kde = len(idx_accept)
            if self.n_kde < 1:
                raise ValueError("Invalid weights")
            elif self.n_kde > self.n_kde_init:
                self.initialise_n_kde()
                idx_accept = self.deweighted_resampling(Y, self.n_kde)
        
        self.Xobs = X[idx_accept]
        self.weights = self.cleansing_weights(Y[idx_accept])
        self.n_kde = self.Xobs.size(0)
        self.set_bandwidth()
        self._compute_covariance()
        if self.compute_cdf:
            self._compute_constant()
        
    def _compute_constant(self):
        p_lb = Phi(self.bounds[0], loc=self.Xobs, covariance_matrix=self.covariance)
        p_ub = Phi(self.bounds[1], loc=self.Xobs, covariance_matrix=self.covariance)
        self.constant = self.standardise_tensor(p_ub - p_lb)

    def set_bandwidth(self):
        """
        Set bandwidth (self.bw) by following the methods (self.bw_method)
        """
        self.neff = 1.0 / (self.weights ** 2).sum()
        if self.bw_method == 'scott':
            self.bw = self.neff.pow(-1./(self.n_dims+4))
        elif self.bw_method == 'silverman':
            self.bw = (self.neff * (self.n_dims+2.0)/4.0).pow(-1./(self.n_dims+4))
        
    def _compute_covariance(self):
        """
        Set covariance matrix (self.covariance)
        """
        # Compute the mean and residuals
        _mean = self.weights @ self.Xobs
        _residual = self.Xobs - _mean.unsqueeze(0)
        # Compute the biased covariance
        self._data_covariance = (_residual.T * self.weights.unsqueeze(0)) @ _residual
        # Correct for bias(http://en.wikipedia.org/wiki/Weighted_arithmetic_mean#Weighted_sample_covariance)
        self._data_covariance /= 1 - self.weights.pow(2).sum()
        self.covariance = self.make_cov_psd(self._data_covariance * self.bw.pow(2))

    def pdf(self, X):
        """
        Compute the probability density function (PDF) at X
        
        Args:
        - X: torch.tensor, the observed data X
        
        
        Return:
        - Ypreds: torch.tensor, the PDF at X
        """
        n_X = len(X)

        x_AA = (
            self.Xobs.repeat(n_X, 1, 1) - X.unsqueeze(1)
        ).reshape(int(self.n_kde*n_X), self.n_dims)

        Npdfs_AA = self.safe_mvn_prob(
            self.zeros(self.n_dims),
            self.covariance,
            x_AA,
        ).reshape(n_X, self.n_kde)
        
        if not self.bounds == None:
            # Thresholding out-of-bound samples
            indices_min = (X < self.bounds[0]).any(axis=1)
            indices_max = (X > self.bounds[1]).any(axis=1)
            Npdfs_AA[indices_min] = self.zeros(self.n_kde)
            Npdfs_AA[indices_max] = self.zeros(self.n_kde)
            
            if self.compute_cdf:
                Ypreds = (self.weights / self.constant) @ Npdfs_AA.T
            else:
                Ypreds = self.weights @ Npdfs_AA.T
        else:
            Ypreds = self.weights @ Npdfs_AA.T
        return Ypreds
    
    def logpdf(self, X):
        """
        Compute the log probability density function (PDF) at X
        
        Args:
        - X: torch.tensor, the observed data X
        
        Return:
        - Ypreds: torch.tensor, the log PDF at X
        """
        return self.pdf(X).log()
    
    def rejection_sampling(self, mean, cov, cnt, n_repeat=10):
        """
        Rejection sampling from truncated Gaussian prior
        
        Args:
        - mean: torch.tensor, the mean vector of Gaussian
        - cov: torch.tensor, the covariance matrix of Gaussian
        - cnt: int, the number of samples
        - n_repeat: int, the number of iteration until len(samples) >= cnt
        
        Return:
        - samples: torch.tensor, the accepted samples from truncated Gaussian prior
        """
        samples = self.null()
        for i in range(n_repeat):
            cov = self.make_cov_psd(cov)
            samples_raw = MultivariateNormal(
                mean,
                cov,
            ).sample(torch.Size([int(n_repeat*cnt)]))
            
            indices_min = (samples_raw < self.bounds[0]).any(axis=1)
            indices_max = (samples_raw > self.bounds[1]).any(axis=1)
            indices_accept = torch.logical_or(indices_min, indices_max).logical_not()
            samples_accepted = samples_raw[indices_accept]
            samples = torch.cat([samples, samples_accepted])
            if (len(samples) >= cnt):
                break
        if (len(samples) > cnt):
            samples = samples[:cnt]
        return samples
    
    def sample_from_Gaussian(self, mean, cov, cnt, n_repeat=10):
        """
        Sampling from each Gaussian
        
        Args:
        - mean: torch.tensor, the mean vector of Gaussian
        - cov: torch.tensor, the covariance matrix of Gaussian
        - cnt: int, the number of samples
        - n_repeat: int, the number of iteration until len(samples) >= cnt
        
        Return
        - samples: torh.tensor, samples from Gaussian
        """
        if cnt == 0:
            warnings.warn("invalid Gaussian in the kernel density estimation")
            return self.null()
        elif (cov == 0).all():
            warnings.warn("invalid Gaussian in the kernel density estimation")
            return self.null()
        else:
            cov = self.make_cov_psd(cov)
            if self.bounds == None:
                samples = MultivariateNormal(
                    mean,
                    cov,
                ).sample(torch.Size([cnt]))
            else:
                samples = self.rejection_sampling(mean, cov, cnt, n_repeat=n_repeat)
            return samples
    
    def sample(self, N_rec):
        """
        Sampling from KDE
        
        Args:
        - N_rec: int, the number of samples
        
        Return
        - samples: torh.tensor, samples from KDE
        """
        cnts = self.weights * N_rec
        cnt_kde = cnts.type(torch.int)
        if cnt_kde.sum() < N_rec:
            cnt_kde = (2 * cnts).type(torch.int)

        samples = torch.cat([
            self.sample_from_Gaussian(
                self.Xobs[i],
                self.covariance,
                cnt,
            )
            for i, cnt in enumerate(cnt_kde)
        ])
        
        if len(samples) > N_rec:
            indice = torch.multinomial(self.ones(len(samples)), N_rec)
            samples = samples[indice]
        return samples



import copy
import torch
import torch.distributions as D
from abc import ABC, abstractmethod
import numpy as np
import pandas as pd
from ._utils import TensorManager
from ._tmvn import TruncatedMVN
from .mvnorm import multivariate_normal_cdf as Phi


class BasePrior(ABC, TensorManager):
    def __init__(self):
        super().__init__()
    
    @abstractmethod
    def sample(self, X):
        r"""Sampling from the prior"""
        pass
    
    @abstractmethod
    def pdf(self, X):
        r"""Return the probability density function of the prior"""
        pass

class Uniform(BasePrior):
    def __init__(self, bounds):
        """
        Uniform prior class
        
        Args:
        - bounds: torch.tensor, the lower and upper bounds for each dimension
        """
        super().__init__() # call TensorManager
        self.bounds = self.standardise_tensor(bounds)
        self.n_dims = self.bounds.shape[1]
        self.type = "continuous"
        
    def sample(self, n_samples, qmc=True):
        """
        Sampling from Uniform prior
        
        Args:
        - n_samples: int, the number of initial samples
        - qmc: bool, sampling from Sobol sequence if True, otherwise simply Monte Carlo sampling.
        
        Return:
        - samples: torch.tensor, the samples from uniform prior
        """
        random_samples = self.rand(self.n_dims, n_samples, qmc=qmc)
        samples = self.bounds[0].unsqueeze(0) + (
            self.bounds[1] - self.bounds[0]
        ).unsqueeze(0) * random_samples
        return samples
    
    def pdf(self, samples):
        """
        The probability density function (PDF) over samples
        
        Args:
        - samples: torch.tensor, the input where to compute PDF
        
        Return:
        - pdfs: torch.tensor, the PDF over samples
        """
        _pdf = self.ones(len(samples)) * (1/(self.bounds[1] - self.bounds[0])).prod()
        _ood = torch.logical_or(
            (samples >= self.bounds[1]).any(axis=1), 
            (samples <= self.bounds[0]).any(axis=1),
        ).logical_not()
        return self.standardise_tensor(_pdf * _ood)
    
    def logpdf(self, samples):
        """
        The log probability density function (PDF) over samples
        
        Args:
        - samples: torch.tensor, the input where to compute PDF
        
        Return:
        - pdfs: torch.tensor, the log PDF over samples
        """
        _logpdf = self.ones(len(samples)) * (1/(self.bounds[1] - self.bounds[0])).prod().log()
        _ood = torch.logical_or(
            (samples >= self.bounds[1]).any(axis=1), 
            (samples <= self.bounds[0]).any(axis=1),
        ).logical_not()
        return self.standardise_tensor(_logpdf * _ood)
    
class Gaussian(BasePrior):
    def __init__(self, mu, cov):
        """
        Gaussian prior class.
        
        Args:
        - mu: torch.tensor, the mean vector of Gaussian distribution
        - cov: torch.tensor, the covariance matrix of Gaussian distribution
        """
        super().__init__() # call TensorManager
        self.n_dims = len(mu)
        self.mvn = D.MultivariateNormal(
            self.standardise_tensor(mu), 
            self.standardise_tensor(cov),
        )
        self.type = "continuous"
    
    def sample(self, n_samples):
        """
        Sampling from Gaussian prior
        
        Args:
        - n_samples: int, the number of initial samples
        
        Return:
        - samples: torch.tensor, the samples from Gaussian prior
        """
        samples = self.mvn.sample(torch.Size([n_samples]))
        return samples
    
    def pdf(self, x):
        """
        The probability density function (PDF) over x.
        
        Args:
        - x: torch.tensor, the input where to compute PDF
        
        Return:
        - pdfs: torch.tensor, the PDF over x
        """
        pdfs = self.mvn.log_prob(x).exp()
        return pdfs
        
class TruncatedGaussian(BasePrior):
    def __init__(self, mu, cov, bounds):
        """
        Truncated Gaussian prior class.
        
        Args:
        - mu: torch.tensor, the mean vector of Gaussian distribution
        - cov: torch.tensor, the covariance matrix of Gaussian distribution
        """
        super().__init__() # call TensorManager
        self.n_dims = len(mu)
        self.mvn = D.MultivariateNormal(
            self.standardise_tensor(mu), 
            self.standardise_tensor(cov),
        )
        self.type = "continuous"
        self.bounds = self.standardise_tensor(bounds)
        p_lb = Phi(bounds[0], loc=mu, covariance_matrix=cov)
        p_ub = Phi(bounds[1], loc=mu, covariance_matrix=cov)
        self.constant = self.standardise_tensor(p_ub - p_lb)
        self.tmvn = TruncatedMVN(mu, cov, bounds)
    
    def sample(self, n_samples):
        """
        Sampling from Gaussian prior
        
        Args:
        - n_samples: int, the number of initial samples
        
        Return:
        - samples: torch.tensor, the samples from Gaussian prior
        """
        samples = self.tmvn.sample(n_samples)
        return self.standardise_tensor(samples)
    
    def pdf(self, x):
        """
        The probability density function (PDF) over x.
        
        Args:
        - x: torch.tensor, the input where to compute PDF
        
        Return:
        - pdfs: torch.tensor, the PDF over x
        """
        pdfs = self.mvn.log_prob(x).exp() / self.constant
        # Thresholding out-of-bound samples
        indices_min = (x < self.bounds[0]).any(axis=1)
        indices_max = (x > self.bounds[1]).any(axis=1)
        pdfs[indices_min] = self.zeros(len(x))[indices_min]
        pdfs[indices_max] = self.zeros(len(x))[indices_max]
        return pdfs

class CategoricalPrior(BasePrior):
    def __init__(self, categories):
        """
        Categorical prior class
        
        Args:
        - categories: list, the number of dimensions x number of categories
        """
        super().__init__() # call TensorManager
        self.type = "categorical"
        self.set_prior(categories)
        
    def set_prior(self, categories):
        """
        Set parameters and functions
        
        Args:
        - categories: list, the number of dimensions x number of categories
        """
        self.categories = [self.tensor(dim) for dim in categories]
        self.n_dims = len(categories)
        self.n_categories = self.tensor(
            [len(category) for category in self.categories]
        ).long()
        self.weights = [self.ones(n_category) * 0.5 for n_category in self.n_categories]
        self.initialise()
        
    def initialise(self):
        """
        Reset the weights of the categorical distribution classes
        """
        self.cats = [D.Categorical(weight) for weight in self.weights]
    
    def find_corresponding_categories(self, indices):
        """
        Find the corresponding categories from the indices
        
        Args:
        - indices: torch.tensor, the indices for each dimension and category
        
        Return:
        - samples: torch.tensor, categories correspinding to the given indices
        """
        indexing_tensor = self.arange(indices.size(1))
        samples = torch.stack([
            self.categories[dim][indices[:, dim]] for dim in indexing_tensor
        ], dim=1)
        return samples
    
    def sample_both(self, n_samples):
        """
        Sampling both categorical values and indices from categorical prior
        
        Args:
        - n_samples: int, the number of samples
        
        Return:
        - samples: torch.tensor, random samples from categorical distribution
        - indices: torch.tensor, indices of random samples
        """
        indices = torch.vstack([cat.sample(torch.Size([n_samples])) for cat in self.cats]).T
        samples = self.find_corresponding_categories(indices)
        return self.standardise_device(samples), indices
    
    def sample(self, n_samples):
        """
        Sampling from categorical prior
        
        Args:
        - n_samples: int, the number of samples
        
        Return:
        - samples: torch.tensor, random samples from categorical distribution
        """
        samples, _ = self.sample_both(n_samples)
        return samples
    
    def logpdf(self, x):
        """
        The log probability mass function (PMF) over x
        
        Args:
        - x: torch.tensor, the input where to compute PDF
        
        Return:
        - pmfs: torch.tensor, the PMF over x
        """
        return torch.vstack([
            self.cats[dim].log_prob(x[:,dim]) for dim in range(self.n_dims)
        ]).sum(axis=0)
    
    def pdf(self, x):
        """
        The probability mass function (PMF) over x
        
        Args:
        - x: torch.tensor, the input where to compute PDF
        
        Return:
        - pmfs: torch.tensor, the PMF over x
        """
        return self.logpdf(x).exp()
    
class BinaryPrior(BasePrior):
    def __init__(self, n_dims):
        """
        Bernoulli (Binary) prior class
        
        Args:
        - n_dims: int, the number of dimensions
        """
        super().__init__() # call TensorManager
        self.n_dims = n_dims
        self.type = "binary"
        self.prior_binary = D.Bernoulli(self.ones(self.n_dims) * 0.5)
        
    def sample(self, n_samples):
        """
        Sampling from Bernoulli prior
        
        Args:
        - n_samples: int, the number of samples
        
        Return:
        - samples: torch.tensor, random samples from Bernoulli distribution
        """
        return self.prior_binary.sample(torch.Size([n_samples]))
    
    def pdf(self, samples):
        """
        The probability mass function (PMF) over samples
        
        Args:
        - samples: torch.tensor, the input where to compute PDF
        
        Return:
        - pmfs: torch.tensor, the PMF over samples
        """
        return self.prior_binary.log_prob(samples).exp().prod(axis=1)
    
    def logpdf(self, samples):
        """
        The log probability mass function (PMF) over samples
        
        Args:
        - samples: torch.tensor, the input where to compute PDF
        
        Return:
        - pmfs: torch.tensor, the log PMF over samples
        """
        return self.prior_binary.log_prob(samples).sum(axis=1)

class MixedBinaryPrior(BasePrior):
    def __init__(
        self, 
        n_dims_cont, 
        n_dims_binary,
        bounds,
        continous_first=True,
    ):
        """
        Mixed prior of Bernoulli and uniform distributions
        
        Args:
        - n_dims_cont: int, the number of dimensions for continuous variables
        - n_dims_binary: int, the number of dimensions for binary variables
        - bounds: torch.tensor, the lower and upper bounds for each dimension only for continuous variables
        - continous_first: bool, [cont, binary] if true, otherwise [binary, cont].
        """
        super().__init__() # call TensorManager
        self.n_dims_cont = n_dims_cont
        self.n_dims_binary = n_dims_binary
        self.bounds = self.standardise_tensor(bounds)
        self.continous_first = continous_first
        self.type = "mixedbinary"
        self.set_prior()
        
    def set_prior(self):
        """
        Set mixed prior
        """
        self.prior_cont = Uniform(self.bounds)
        self.prior_binary = BinaryPrior(self.n_dims_binary)
        
    def separate_samples(self, x):
        """
        Separate mixed variables to each type
        
        Args:
        - x: torch.tensor, the input where to compute PDF
        
        Return:
        - x_cont: torch.tensor, continuous variables
        - x_binary: torch.tensor, binary variables
        """
        if self.continous_first:
            x_cont = x[:, :self.n_dims_cont]
            x_binary = x[:, self.n_dims_cont:]
        else:
            x_binary = x[:, :self.n_dims_binary]
            x_cont = x[:, self.n_dims_binary:]
        return x_cont, x_binary
        
    def sample(self, n_samples):
        """
        Sampling from mixed prior
        
        Args:
        - n_samples: int, the number of samples
        
        Return:
        - samples: torch.tensor, random samples from mixed distribution
        """
        samples_cont = self.prior_cont.sample(n_samples)
        samples_binary = self.prior_binary.sample(n_samples)
        if self.continous_first:
            return torch.hstack([samples_cont, samples_binary])
        else:
            return torch.hstack([samples_binary, samples_cont])
    
    def pdf(self, x):
        """
        The probability density function (PDF) over x
        
        Args:
        - x: torch.tensor, the input where to compute PDF
        
        Return:
        - pdfs: torch.tensor, the PDF over samples
        """
        x_cont, x_binary = self.separate_samples(x)
        pdf_cont = self.prior_cont.pdf(x_cont)
        pdf_binary = self.prior_binary.pdf(x_binary)
        return pdf_cont * pdf_binary
    
    def logpdf(self, x):
        """
        The log probability density function (PDF) over x
        
        Args:
        - x: torch.tensor, the input where to compute PDF
        
        Return:
        - pdfs: torch.tensor, the log PDF over samples
        """
        x_cont, x_binary = self.separate_samples(x)
        pdf_cont = self.prior_cont.logpdf(x_cont)
        pdf_binary = self.prior_binary.logpdf(x_binary)
        return pdf_cont + pdf_binary

class MixedCategoricalPrior(BasePrior):
    def __init__(self, n_dims_cont, n_dims_disc, categories, bounds, continous_first=True):
        """
        Mixed prior of categorical and uniform distributions
        
        Args:
        - n_dims_cont: int, the number of dimensions for continuous variables
        - n_dims_binary: int, the number of dimensions for categorical variables
        - n_discrete: int, the number of categories for each dimension
        - bounds: torch.tensor, the lower and upper bounds for each dimension only for continuous variables
        - categories: torch.tensor, the categories to select for categorical variables.
        - continous_first: bool, [cont, categorical] if true, otherwise [categorical, cont].
        """
        super().__init__() # call TensorManager
        self.n_dims_cont = n_dims_cont
        self.n_dims_disc = n_dims_disc
        self.categories = self.standardise_tensor(categories)
        self.bounds = self.standardise_tensor(bounds)
        self.continous_first = continous_first
        self.type = "mixedcategorical"
        self.set_prior()
        
    def set_prior(self):
        """
        Set mixed prior
        """
        self.prior_cont = Uniform(self.bounds)
        self.prior_disc = CategoricalPrior(self.categories)
        
    def separate_samples(self, x):
        """
        Separate mixed variables to each type
        
        Args:
        - x: torch.tensor, the input where to compute PDF
        
        Return:
        - x_cont: torch.tensor, continuous variables
        - x_disc: torch.tensor, categorical variables
        """
        if self.continous_first:
            x_cont = x[:, :self.n_dims_cont]
            x_disc = x[:, self.n_dims_cont:]
        else:
            x_disc = x[:, :self.n_dims_disc]
            x_cont = x[:, self.n_dims_disc:]
        return x_cont, x_disc
        
    def sample(self, n_samples):
        """
        Sampling from mixed prior
        
        Args:
        - n_samples: int, the number of samples
        
        Return:
        - samples: torch.tensor, random samples from mixed distribution
        """
        samples_cont = self.prior_cont.sample(n_samples)
        samples_disc = self.prior_disc.sample(n_samples)
        if self.continous_first:
            return torch.hstack([samples_cont, samples_disc])
        else:
            return torch.hstack([samples_disc, samples_cont])
    
    def sample_both(self, n_samples):
        """
        Drawing both samples and indices from mixed prior
        
        Args:
        - n_samples: int, the number of samples
        
        Return:
        - samples: torch.tensor, random samples from categorical distribution
        - indices: torch.tensor, indices of categorical samples
        """
        samples_cont = self.prior_cont.sample(n_samples)
        samples_disc, indices = self.prior_disc.sample_both(n_samples)
        if self.continous_first:
            return (
                torch.hstack([samples_cont, samples_disc]), 
                torch.hstack([samples_cont, indices]),
            )
        else:
            return (
                torch.hstack([samples_disc, samples_cont]), 
                torch.hstack([indices, samples_cont]),
            )
    
    def pdf(self, x):
        """
        The probability density function (PDF) over x
        
        Args:
        - x: torch.tensor, the input where to compute PDF
        
        Return:
        - pdfs: torch.tensor, the PDF over samples
        """
        x_cont, x_disc = self.separate_samples(x)
        pdf_cont = self.prior_cont.pdf(x_cont)
        pdf_disc = self.prior_disc.pdf(x_disc)
        return pdf_cont * pdf_disc

class DatasetPrior(BasePrior):
    def __init__(
        self,
        features,
        true_targets,
    ):
        """
        Dataset prior for which all list of possible candidates are given as dataset
        
        Args:
        - features: torch.tensor, the binary inputs
        - true_targets: torch.tensor, the objective to maximize
        """
        super().__init__() # call TensorManager
        self.available_index = self.arange(len(features))
        self.features = self.standardise_tensor(features)
        self.true_targets = self.standardise_tensor(true_targets)
        self.reset_indices(self.available_index)
        self.type = "dataset"
        
    def reset_indices(self, available_index):
        """
        Reset the indices of dataset
        
        Args:
        - available_index: torch.tensor, the available indices that the queried indices are removed.
        """
        self.n_available = available_index.shape[0]
        self.features = self.features[available_index]
        self.true_targets = self.true_targets[available_index]
        self.available_index = self.arange(self.n_available)
        
    def set_substract(self, A, B):
        """
        Substracting the set B from the set A, where len(A) > len(B)
        
        Args:
        - A: torch.tensor, the set of indices.
        - B: torch.tensor, the set of indices.
        
        Return:
        - A-B: torch.tensor, the substracted set (A/B)
        """
        mask = self.ones(A.shape).to(torch.bool)
        mask[B] = 0
        return torch.masked_select(A, mask)
        
    def remove_sampled_index(self, idx_sampled):
        """
        Remove the sampled indices
        
        Args:
        - idx_sampled: torch.tensor, the sampled indices from the available indices.
        """
        available_index = self.set_substract(self.available_index, idx_sampled)
        self.reset_indices(available_index)
    
    def query(self, idx_cand):
        """
        Query Y at given indices.
        Then, update the internal dataset to delete the drawn samples.
        
        Args:
        - X_cand: torch.tensor, the features to query.
        
        Return:
        - Y: torch.tensor, the true values
        """
        Y = self.true_targets[idx_cand]
        self.remove_sampled_index(idx_cand)
        return Y
    
    def sample(self, n_sample):
        """
        Sample both X and Y with the size of n_sample.
        Then, update the internal dataset to delete the drawn samples. 
        
        Args:
        - n_sample: int, the number of samples.
        
        Return:
        - X: torch.tensor, the features.
        - Y: torch.tensor, the true values.
        """
        idx_sampled = self.randperm(self.n_available)[:n_sample]
        X = self.features[idx_sampled]
        Y = self.true_targets[idx_sampled]
        self.remove_sampled_index(idx_sampled)
        return X, Y
    
    def sample_feature(self, n_sample):
        """
        Sample X with the size of n_sample
        
        Args:
        - n_sample: int, the number of samples.
        
        Return:
        - X: torch.tensor, the features.
        """
        idx_sampled = self.randperm(self.n_available)[:n_sample]
        X = self.features[idx_sampled]
        return idx_sampled, X
    
    def available_candidates(self):
        """
        Sample all available X
        
        Return:
        - X: torch.tensor, the features.
        """
        return self.features
    
    def pdf(self, X):
        return self.ones(len(X)) / len(X)



"""
    - 1. Integrate multivariate normal density (CDFs)
    - 2. Easily obtain partial derivatives of CDFs w.r.t location, mean and covariance (and higher derivatives)
    - 3. Manipulate quantities within a tensor-based framework (e.g. broadcasting is fully supported)

"""

from .multivariate_normal_cdf import multivariate_normal_cdf
from .integration import integration


import copy
import torch
import gpytorch
from torch.distributions.normal import Normal
from torch.distributions.multivariate_normal import MultivariateNormal
from ._scale_vbq import ScaleVanillaGP
from .._gp import ExactGPModel
from .._utils import Utils
from .._kernel import Kernel
from .._rchq import recombination
from .._weights import WeightsStabiliser

class ManagingGPHyperparameters:
    def __init__(self, rng=100):
        """
        Managing Gaussian process hyperparameter
        
        Args:
        - rng: int, the range of the likelihood bounds
        """
        self.rng = rng
        
    def show_hypers(self, model):
        """
        Show the current hyperparameter values
        
        Args:
        - model: gpytorch.models, the Gaussian process model
        """
        print(
            "lik_var: " + str(model.likelihood.noise.item()) + '\n' +
            "lengthscale: " + str(model.covar_module.base_kernel.lengthscale.item())+ '\n' +
            "outputscale: " + str(model.covar_module.outputscale.item())
        )
        
    def extract_hypers(self, model):
        """
        Return the current hyperparameters.
        
        Args:
        - model: gpytorch.models, the Gaussian process model
        
        Return:
        - Theta: torch.tensor, the current hyperparameters
        """
        return torch.tensor([
            model.likelihood.noise.detach(),
            model.covar_module.base_kernel.lengthscale.detach(),
            model.covar_module.outputscale.detach(),
        ])

    def set_hypers(self, model, theta):
        """
        Update the GP hyperparameters.
        
        Args:
        - model: gpytorch.models, the Gaussian process model
        - theta: torch.tensor, the hyperparameters to set
        
        Return:
        - model: gpytorch.models, the updated Gaussian process model
        """
        hypers = {
            'likelihood.noise_covar.noise': theta[0],
            'covar_module.base_kernel.lengthscale': theta[1],
            'covar_module.outputscale': theta[2],
        }
        model.initialize(**hypers)
        return model

    def reset_GP(self, train_x, train_y, theta):
        """
        Reset the GP model.
        
        Args:
        - train_x: torch.tensor, the input
        - train_y: torch.tensor, the output
        - theta: torch.tensor, the hyperparameters to set
        
        Return:
        - model: gpytorch.models, the updated Gaussian process model
        """
        likelihood = gpytorch.likelihoods.GaussianLikelihood()
        likelihood.noise_covar.register_constraint(
            "raw_noise",
            gpytorch.constraints.Interval(theta[0] / self.rng, theta[0] * self.rng)
        )
        gp_kernel = gpytorch.kernels.ScaleKernel(gpytorch.kernels.RBFKernel())
        model = ExactGPModel(train_x, train_y, likelihood, gp_kernel)
        model = self.set_hypers(model, theta)
        return model
    
class LogMarginalLikelihood(ManagingGPHyperparameters):
    def __init__(self, gp):
        """
        Log marginal likelihood of FITBO model
        
        Args:
        - gp: FitboGP class, the function of underlying model
        """
        super().__init__(rng=100) # inherit ManagingGPHyperparameters class
        device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
        self.utils = Utils(device)
        self.eps = -torch.sqrt(torch.tensor(torch.finfo().max))
        self.Xobs = copy.deepcopy(gp.model.train_inputs[0])
        self.ymax = copy.deepcopy(gp.model.train_targets.max())
        self.eta = copy.deepcopy(gp.alpha)
        self.fobs = gp.Y_unwarp #(gp.Y_unwarp - gp.Y_unwarp.mean()) / gp.Y_unwarp.std()
        self.n_data = self.Xobs.size(0)
        self.theta_map = self.extract_hypers(gp.model)
        
    def log_to_exp_transform(self, theta):
        """
        Transform back the hyperparameters to the original space
        
        Args:
        - theta: torch.tensor, the log-transformed hyperparameters
        """
        Theta = theta.exp()
        if Theta.dim() == 1:
            Theta[0] = self.eta + Theta[0]
        else:
            Theta[:,0] = self.eta + Theta[:,0]
        return Theta

    def mll(self, Theta):
        """
        Compute the marginal log likelihood of FITBO model
        
        Args:
        - Theta: torch.tensor, the hyperparameters
        
        Return:
        - mll: torch.tensor, the marginal log likelihood
        """
        # hypers
        eta = Theta[0]
        theta = Theta[1:]

        # reset hypers
        gobs_pseudo = eta.sign() * (2*(eta - self.fobs)).sqrt()
        model = self.reset_GP(self.Xobs, gobs_pseudo, theta)
        model.eval()

        # g space prediction
        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            function_dist = model(self.Xobs)
            output = model.likelihood(function_dist)
        mu_g = output.loc
        var_g = output.variance
        covar_g = output.covariance_matrix

        # f space prediction
        mu_f = eta - 0.5 * (mu_g**2 + var_g)
        covar_f = mu_g.unsqueeze(1) * covar_g * mu_g.unsqueeze(0) + 0.5 * (covar_g ** 2)
        res = MultivariateNormal(
            mu_f.detach(), 
            self.utils.make_cov_psd(covar_f),
        ).log_prob(self.fobs)
        mll = res.div_(self.n_data)
        return mll
    
    def __call__(self, theta):
        """
        Compute the marginal log likelihood at given log-transformed hyperparamters
        
        Args:
        - theta: torch.tensor, the log-transformed hyperparameters
        
        Return:
        - mll: torch.tensor, the marginal log likelihood
        """
        Theta = self.log_to_exp_transform(theta)
        try:
            return self.mll(Theta)
        except:
            return self.eps

def sampling_hypers(model, hyperprior, n_hypers=1000, use_map=False):
    """
    Sampling hypersamples with fully Bayesian Gaussian process (FBGP) model

    Args:
    - model: FitboGP class, the function of underlying model
    - hyperprior: RBFHyperPrior class, the function of hyperprior
    - n_hypers: int, the number of hypersamples from hyperprior
    - use_map: bool, use MAP estimated hypersamples as the mean if true, otherwise not.

    Return:
    - Hypersamples: torch.tensor, the hypersamples from hyperprior
    - LMLs: torch.tensor, the computed log marginal likelihoods
    """
    lml = LogMarginalLikelihood(model)
    if use_map:
        hyperprior.initialise(lml.theta_map)
    hypersamples = hyperprior.sample(n_hypers)
    hypersamples = torch.vstack([
        torch.cat([torch.tensor([-10]), lml.theta_map.log()]),
        hypersamples,
    ])
    LMLs = torch.tensor([lml(theta) for theta in hypersamples])
    Hypersamples = lml.log_to_exp_transform(hypersamples)
    return Hypersamples, LMLs

def quadrature_distillation(Hypersamples, LMLs, gp_kernel, n_nys=100, n_qd=50):
    """
    Quadrature distillation for sparsifying weighted hypersamples

    Args:
    - Hypersamples: torch.tensor, the hypersamples from hyperprior
    - LMLs: torch.tensor, the computed log marginal likelihoods
    - gp_kernel: gpytorch.kernels, the function of kernel for Bayesian quadrature model
    - n_nys: int, the number of hypersamples for Nyström approximation
    - n_qd: int, the number of resulting distilled hypersamples

    Return:
    - w_qd: torch.tensor, the non-negative weights for distilled hypersamples
    - Theta_qd: torch.tensor, the distilled hypersamples
    """
    device = torch.device('cuda') if torch.cuda.is_available() else torch.device('cpu')
    
    # hyperposterior as weighted hyperprior samples
    n_rec = len(Hypersamples)
    weights = (LMLs - LMLs.max()).exp()
    
    W = WeightsStabiliser()
    weights = W.cleansing_weights(weights)
    idx_nys = W.deweighted_resampling(weights, n_nys)
    Hyper_nys = Hypersamples[idx_nys]
    
    # modelling hyper-GP
    VBQ = ScaleVanillaGP(Hypersamples, LMLs, gp_kernel, device)
    kernel = Kernel(VBQ.model, mode="kernel")
    
    # recombination
    idx, w_qd = recombination(
        Hypersamples,
        Hyper_nys,
        n_qd,
        kernel,
        device,
        init_weights=weights,
    )
    Theta_qd = Hypersamples[idx]
    return w_qd, Theta_qd

class FullyBayesianGP(LogMarginalLikelihood):
    def __init__(self, gp, w_qd, Theta_qd):
        """
        Fully Bayesian Gaussian process model

        Args:
        - gp: FitboGP class, the function of underlying model
        - w_qd: torch.tensor, the non-negative weights for distilled hypersamples
        - Theta_qd: torch.tensor, the distilled hypersamples
        """
        super().__init__(gp) # inherit LogMarginalLikelihood class
        self.w_qd = w_qd.detach()
        self.Theta_qd = Theta_qd.detach()
        self.is_fbgp = True
    
    def fitbo_predict(self, x_test, Theta):
        """
        Posterior predictive distribution from FITBO GP model

        Args:
        - x_test: torch.tensor, the input
        - Theta: torch.tensor, the set of hypersamples
        
        Return:
        - mu_f: torch.tensor, the posterior predictive mean
        - var_f: torch.tensor, the posterior predictive variance
        """
        gobs_pseudo = Theta[0].sign() * (2*(Theta[0] - self.fobs)).sqrt()
        model = self.reset_GP(self.Xobs, gobs_pseudo, Theta[1:])
        
        # g space prediction
        model.eval()
        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            function_dist = model(x_test)
            output = model.likelihood(function_dist)
        mu_g = output.loc
        var_g = output.variance
        var_g[var_g < 0] = 0

        # f space prediction
        mu_f = Theta[0] - 0.5 * (mu_g**2 + var_g)
        var_f = mu_g * var_g * mu_g + 0.5 * (var_g ** 2)
        return mu_f, var_f

    def fitbo_predict_batch(self, x_test, Theta):
        """
        A batch of posterior predictive distribution from FITBO GP model

        Args:
        - x_test: torch.tensor, the input
        - Theta: torch.tensor, the set of hypersamples
        
        Return:
        - mu_batch: torch.tensor, the batch of posterior predictive mean
        - var_batch: torch.tensor, the batch of  posterior predictive variance
        """
        mu, var = self.fitbo_predict(x_test, Theta)
        var[var < 0] = 0
        return torch.stack([mu, var])

    def batch_predict(self, x_test):
        """
        A batch of posterior predictive distribution from FITBO GP model

        Args:
        - x_test: torch.tensor, the input
        
        Return:
        - mu_batch: torch.tensor, the batch of posterior predictive mean
        - var_batch: torch.tensor, the batch of  posterior predictive variance
        """
        batch_pred = torch.stack([
        self.fitbo_predict_batch(x_test, Theta)
            for Theta in self.Theta_qd
        ])
        mu_batch, var_batch = batch_pred[:,0,:], batch_pred[:,1,:]
        return mu_batch, var_batch
    
    def marginal_predict(self, x_test):
        """
        Marginal posterior predictive distribution from FITBO GP model

        Args:
        - x_test: torch.tensor, the input
        
        Return:
        - mu_fbgp: torch.tensor, the marginal posterior predictive mean
        - var_fbgp: torch.tensor, the marginal posterior predictive variance
        """
        mu_batch, var_batch = self.batch_predict(x_test)
        mu_fbgp = self.w_qd @ mu_batch
        var_fbgp = self.w_qd @ (var_batch + mu_batch.pow(2)) - (self.w_qd @ mu_batch).pow(2)
        return mu_fbgp, var_fbgp
    
    def marginal_predictive_mean(self, x_test):
        """
        Marginal posterior predictive mean from FITBO GP model

        Args:
        - x_test: torch.tensor, the input
        
        Return:
        - mu_fbgp: torch.tensor, the marginal posterior predictive mean
        """
        mu_batch, _ = self.batch_predict(x_test)
        return self.w_qd @ mu_batch
    
    def marginal_predictive_covariance(self, x_test, y_test):
        """
        Marginal posterior predictive covariance from FITBO GP model

        Args:
        - x_test: torch.tensor, the input
        - y_test: torch.tensor, the input
        
        Return:
        - cov_ny: torch.tensor, the marginal posterior predictive covariance
        """
        mu_x, _ = self.batch_predict(x_test)
        mu_y, _ = self.batch_predict(y_test)
        Ex = self.w_qd @ mu_x
        Ey = self.w_qd @ mu_y
        W = 1/(1 - self.w_qd.pow(2).sum())
        cov_xy = W * (self.w_qd.unsqueeze(1) * (mu_x - Ex.unsqueeze(0))).T @ (mu_y - Ey.unsqueeze(0))
        return cov_xy



import copy
import torch
from .._gp import update_gp, predict, predictive_covariance
from .._utils import Utils


class ScaleMmltGP(Utils):
    def __init__(
        self,
        Xobs,
        Yobs,
        gp_kernel,
        label="wsabim",
        alpha_factor=1,
        lik=1e-10,
        training_iter=10000,
        thresh=0.01,
        lr=0.1,
        rng=10,
        train_lik=False,
        optimiser="BoTorch",
    ):
        """
        Scale MMLT BQ modelling
        Scale MMLT is a doubly warped GP modelling. It consists of three levels of warped space.
        The observation y is assumed to belong to g space. (query returns true log likelihoods.)
        WSABI modelling permits a factorisation trick in log-space.
         _________________________ _______________________________________ __________________
        |          f space        |                g space                |     h space      |
        |_________________________|_______________________________________|__________________|
        |             f           |               f / exp(β)              |  h = log(g + 1)  |
        |        f = g exp(β)     |             g = exp(h) - 1            |        h         |
        |_________________________|_______________________________________|__________________|
        |      f = GP(μ_f, σ_f)   |           g = GP(μ_g, σ_g)            | h = GP(μ_h, σ_h) |
        |     μ_f = μ_g(x)exp(β)  |        μ_g = exp(μ_h + 1/2 σ_h) - 1   |        μ_h       |
        |   σ_f = σ_g(x,y)exp(2β) | σ_g = μ_g(x)μ_g(y)(exp(σ_h(x,y)) - 1) |        σ_h       |
        |_________________________|_______________________________________|__________________|

        ScaleMmltGP class summarises the functions of training, updating the warped GP model.
        This also provides the prediction and kernel of WSABI GP.
        The modelling of WSABI-L and WSABI-M can be easily switched by changing "label".
        The above table is the case of WSABI-M.

        Args:
           - Xobs: torch.tensor, X samples, X belongs to prior measure.
           - Yobs: torch.tensor, Y observations, Y = true_likelihood(X).
           - gp_kernel: gpytorch.kernels, GP kernel function
           - device: torch.device, device, cpu or cuda
           - label: string, the wsabi type, ["wsabil", "wsabim"]
           - lik: float, the initial value of GP likelihood noise variance
           - train_iter: int, the maximum iteration for GP hyperparameter training.
           - thresh: float, the threshold as a stopping criterion of GP hyperparameter training.
           - lr: float, the learning rate of Adam optimiser
           - rng: int, tne range coefficient of GP likelihood noise variance
           - train_like: bool, flag whether or not to update GP likelihood noise variance
           - optimiser: string, select the optimiser ["L-BFGS-B", "Adam"]
        """
        super().__init__()
        self.gp_kernel = gp_kernel
        self.alpha_factor = 1
        self.alpha = alpha_factor
        self.lik = lik
        self.training_iter = training_iter
        self.thresh = thresh
        self.lr = lr
        self.rng = rng
        self.train_lik = train_lik
        self.optimiser = optimiser
        self.is_bq = True

        self.jitter = self.tensor(0)  # 1e-6
        self.Y_log = copy.deepcopy(Yobs)

        self.model = update_gp(
            Xobs,
            self.process_y_warping_with_scaling(Yobs),
            gp_kernel,
            self.device,
            lik=self.lik,
            training_iter=self.training_iter,
            thresh=self.thresh,
            lr=self.lr,
            rng=self.rng,
            train_lik=self.train_lik,
            optimiser=self.optimiser,
        )

    def process_y_warping_with_scaling(self, y_obs):
        """
        Args:
           - y_obs: torch.tensor, observations of true_loglikelihood

        Returns:
           - y_h: torch.tensor, warped observations in h space that contains no anomalies and the updated alpha hyperparameter.
        """

        y = self.remove_anomalies(y_obs)
        self.beta = torch.max(y)
        y_g = torch.exp(y - self.beta)
        y_h = self.warp_from_g_to_h(y_g)
        return y_h

    def warp_from_g_to_h(self, y_g):
        """
        Args:
           - y_f: torch.tensor, warped observations in g space

        Returns:
           - y_g: torch.tensor, warped observations in h space
        """
        y_h = torch.log(y_g + 1)
        return y_h

    def unwarp_from_h_to_g(self, y_h):
        """
        Args:
           - y_f: torch.tensor, warped observations in h space

        Returns:
           - y_g: torch.tensor, warped observations in g space
        """
        y_g = torch.exp(y_h) - 1
        return y_g
    

    def cat_observations_with_scaling(self, X, Y):
        """
        Args:
           - X: torch.tensor, X samples to be added to the existing data Xobs
           - Y: torch.tensor, unwarped Y observations to be added to the existing data Yobs

        Returns:
           - Xall: torch.tensor, X samples that contains all samples
           - Yall: torch.tensor, warped Y observations that contains all observations
        """
        Xobs = self.model.train_inputs[0]
        Yobs_log = copy.deepcopy(self.Y_log)
        if len(self.model.train_targets.shape) == 0:
            Yobs_log = Yobs_log.unsqueeze(0)
        Xall = torch.cat([Xobs, X])
        Yall_log = torch.cat([Yobs_log, Y])
        self.Y_log = copy.deepcopy(Yall_log)
        Yall_h = self.process_y_warping_with_scaling(Yall_log)
        return Xall, Yall_h

    def update_mmlt_gp_with_scaling(self, X, Y):
        """
        Args:
           - X: torch.tensor, X samples to be added to the existing data Xobs
           - Y: torch.tensor, unwarped Y observations to be added to the existing data Yobs
        """
        X_h, Y_h = self.cat_observations_with_scaling(X, Y)
        self.model = update_gp(
            X_h,
            Y_h,
            self.gp_kernel,
            self.device,
            lik=self.lik,
            training_iter=self.training_iter,
            thresh=self.thresh,
            lr=self.lr,
            rng=self.rng,
            train_lik=self.train_lik,
            optimiser=self.optimiser,
        )

    def retrain_gp_with_scaling(self):
        X_h = self.model.train_inputs[0]
        Y_h = self.process_y_warping_with_scaling(copy.deepcopy(self.Y_log))
        self.model = update_gp(
            X_h,
            Y_h,
            self.gp_kernel,
            self.device,
            lik=self.lik,
            training_iter=self.training_iter,
            thresh=self.thresh,
            lr=self.lr,
            rng=self.rng,
            train_lik=self.train_lik,
            optimiser=self.optimiser,
        )
    
    def memorise_parameters(self):
        self.likelihood_memory = copy.deepcopy(torch.tensor(self.model.likelihood.noise.item()))
        self.outputsacle_memory = copy.deepcopy(torch.tensor(self.model.covar_module.outputscale.item()))
        self.lengthscale_memory = copy.deepcopy(torch.tensor(self.model.covar_module.base_kernel.lengthscale.item()))

    def remind_parameters(self):
        hypers = {
            'likelihood.noise_covar.noise': self.likelihood_memory,
            'covar_module.outputscale': self.outputsacle_memory,
            'covar_module.base_kernel.lengthscale': self.lengthscale_memory,
        }
        self.model.initialize(**hypers)

    def hspace_predict(self, x):
        """
        Args:
           - x: torch.tensor, x locations to be predicted

        Returns:
           - mu_h: torch.tensor, unwarped predictive mean in h space at given locations x.
           - var_h: torch.tensor, unwarped predictive variance in h space at given locations x.
        """
        mu_h, var_h = predict(x, self.model)
        return mu_h, var_h

    def gspace_predict(self, x):
        """
        Args:
           - x: torch.tensor, x locations to be predicted

        Returns:
           - mu_g: torch.tensor, unwarped predictive mean in g space at given locations x.
           - var_g: torch.tensor, unwarped predictive variance in g space at given locations x.
        """
        mu_h, var_h = self.hspace_predict(x)
        mu_g = (mu_h + 0.5 * var_h).exp() - 1
        var_g = (mu_g ** 2) * (var_h.exp() - 1)
        return mu_g, var_g

    def hspace_mean_predict(self, x):
        """
        Args:
           - x: torch.tensor, x locations to be predicted

        Returns:
           - mu_h: torch.tensor, unwarped predictive mean in h space at given locations x.
        """
        mu_h, _ = self.hspace_predict(x)
        return mu_h

    def gspace_mean_predict(self, x):
        """
        Args:
           - x: torch.tensor, x locations to be predicted

        Returns:
           - mu_g: torch.tensor, unwarped predictive mean in g space at given locations x.
        """
        mu_g, _ = self.gspace_predict(x)
        return mu_g

    def hspace_kernel(self, x, y):
        """
        Args:
           - x: torch.tensor, x locations to be predicted
           - y: torch.tensor, y locations to be predicted

        Args:
           - CLy: torch.tensor, the positive semi-definite Gram matrix of predictive variance in hscape
        """
        return predictive_covariance(x, y, self.model)

    def gspace_kernel(self, x, y):
        """
        Args:
           - x: torch.tensor, x locations to be predicted
           - y: torch.tensor, y locations to be predicted

        Returns:
           - CLy: torch.tensor, the positive semi-definite Gram matrix of predictive variance in gscape
        """
        mu_g_x = self.gspace_mean_predict(x)
        mu_g_y = self.gspace_mean_predict(y)
        cov_h_xy = self.hspace_kernel(x, y)
        if len(cov_h_xy.shape) == 2:
            CLy = mu_g_x.unsqueeze(1) * mu_g_y.unsqueeze(0) * (cov_h_xy.exp() - 1)
        elif len(cov_h_xy.shape) == 3:
            CLy = mu_g_x.unsqueeze(1).unsqueeze(0) * mu_g_y.unsqueeze(1) * (cov_h_xy.exp() - 1)

        d = min(len(x), len(y))
        CLy[range(d), range(d)] = CLy[range(d), range(d)] + self.jitter
        return CLy



import torch
import gpytorch
import numpy as np
import pandas as pd
from botorch.models import SingleTaskGP
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.means import ConstantMean
from gpytorch.kernels import ScaleKernel
from gpytorch.distributions import MultivariateNormal


def default_postprocess_script(x):
    return x

def batch_tanimoto_sim(
    x1: torch.Tensor, x2: torch.Tensor, eps: float = 1e-6
) -> torch.Tensor:
    # Tanimoto distance is proportional to (<x, y>) / (||x||^2 + ||y||^2 - <x, y>) where x and y are bit vectors
    assert x1.ndim >= 2 and x2.ndim >= 2
    dot_prod = torch.matmul(x1, torch.transpose(x2, -1, -2))
    x1_sum = torch.sum(x1**2, dim=-1, keepdims=True)
    x2_sum = torch.sum(x2**2, dim=-1, keepdims=True)
    return (dot_prod + eps) / (
        eps + x1_sum + torch.transpose(x2_sum, -1, -2) - dot_prod
    )


class BitDistance(torch.nn.Module):
    def __init__(self, postprocess_script=default_postprocess_script):
        super().__init__()
        self._postprocess = postprocess_script

    def _sim(self, x1, x2, postprocess, x1_eq_x2=False, metric="tanimoto"):
        # Branch for Tanimoto metric
        if metric == "tanimoto":
            res = batch_tanimoto_sim(x1, x2)
            res.clamp_min_(0)  # zero out negative values
            return self._postprocess(res) if postprocess else res
        else:
            raise RuntimeError(
                "Similarity metric not supported. Available options are 'tanimoto'"
            )


class BitKernel(gpytorch.kernels.Kernel):
    def __init__(self, metric="", **kwargs):
        super().__init__(**kwargs)
        self.metric = metric

    def forward(self, x1, x2, **params):
        return self.covar_dist(x1, x2, **params)

    def covar_dist(
        self,
        x1,
        x2,
        last_dim_is_batch=False,
        dist_postprocess_func=default_postprocess_script,
        postprocess=True,
        **params,
    ):
        if last_dim_is_batch:
            x1 = x1.transpose(-1, -2).unsqueeze(-1)
            x2 = x2.transpose(-1, -2).unsqueeze(-1)

        x1_eq_x2 = torch.equal(x1, x2)

        # torch scripts expect tensors
        postprocess = torch.tensor(postprocess)

        res = None

        # Cache the Distance object or else JIT will recompile every time
        if (
            not self.distance_module
            or self.distance_module._postprocess != dist_postprocess_func
        ):
            self.distance_module = BitDistance(dist_postprocess_func)

        res = self.distance_module._sim(
            x1, x2, postprocess, x1_eq_x2, self.metric
        )

        return res

class TanimotoKernel(BitKernel):
    is_stationary = False
    has_lengthscale = False

    def __init__(self, **kwargs):
        super(TanimotoKernel, self).__init__(**kwargs)
        self.metric = "tanimoto"

    def forward(self, x1, x2, diag=False, **params):
        if diag:
            assert x1.size() == x2.size() and torch.equal(x1, x2)
            return torch.ones(
                *x1.shape[:-2], x1.shape[-2], dtype=x1.dtype, device=x1.device
            )
        else:
            return self.covar_dist(x1, x2, **params)
        
class TanimotoGP(SingleTaskGP):
    def __init__(self, train_X, train_Y):
        super().__init__(train_X, train_Y, GaussianLikelihood())
        self.mean_module = ConstantMean()
        self.covar_module = ScaleKernel(base_kernel=TanimotoKernel())
        self.to(train_X)  # make sure we're on the right device/dtype

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return MultivariateNormal(mean_x, covar_x)
